import pytest

from qbittorrentapi.sync import SyncMainDataDictionary
from qbittorrentapi.sync import SyncTorrentPeersDictionary


@pytest.mark.parametrize("rid", (None, 0, 1, 100000))
def test_maindata1(client, rid):
    assert isinstance(client.sync_maindata(rid=rid), SyncMainDataDictionary)


@pytest.mark.parametrize("rid", (None, 0, 1, 100000))
def test_maindata2(client, rid):
    assert isinstance(client.sync.maindata(rid=rid), SyncMainDataDictionary)
    assert isinstance(client.sync.maindata.delta(), SyncMainDataDictionary)
    assert isinstance(client.sync.maindata.delta(), SyncMainDataDictionary)


def test_maindata3(client):
    client.sync.maindata()
    assert client.sync.maindata._rid != 0
    client.sync.maindata.reset_rid()
    assert client.sync.maindata._rid == 0


@pytest.mark.parametrize("rid", (None, 0, 1, 100000))
def test_torrent_peers(client, rid, orig_torrent):
    assert isinstance(
        client.sync_torrent_peers(rid=rid, torrent_hash=orig_torrent.hash),
        SyncTorrentPeersDictionary,
    )
    assert isinstance(
        client.sync.torrent_peers(rid=rid, torrent_hash=orig_torrent.hash),
        SyncTorrentPeersDictionary,
    )

    assert isinstance(
        client.sync.torrent_peers.delta(torrent_hash=orig_torrent.hash),
        SyncTorrentPeersDictionary,
    )

    client.sync.torrent_peers.reset_rid()
    assert client.sync.torrent_peers._rid == 0



import logging
from json import dumps

import pytest

from qbittorrentapi import APIError
from qbittorrentapi import Client
from qbittorrentapi._attrdict import AttrDict
from qbittorrentapi.decorators import endpoint_introduced
from qbittorrentapi.decorators import handle_hashes
from qbittorrentapi.decorators import response_json
from qbittorrentapi.decorators import response_text
from qbittorrentapi.decorators import version_removed
from qbittorrentapi.request import Request

list2str = Request._list2string


def test_login_required(caplog, app_version):
    client = Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    with caplog.at_level(logging.DEBUG, logger="qbittorrentapi"):
        qbt_version = client.app.version
    assert "Not logged in...attempting login" in caplog.text
    assert qbt_version == app_version

    client.auth_log_out()
    with caplog.at_level(logging.DEBUG, logger="qbittorrentapi"):
        qbt_version = client.app.version
    assert "Not logged in...attempting login" in caplog.text
    assert qbt_version == app_version


@pytest.mark.parametrize("test_hash", (None, "", "xcvbxcvbxcvb"))
@pytest.mark.parametrize("other_param", (None, "", "irivgjkd"))
def test_hash_handler_single_hash(test_hash, other_param):
    class FakeClient:
        @handle_hashes
        def single_torrent_func(self, torrent_hash=None, param=None):
            assert torrent_hash == test_hash
            assert param == other_param

    FakeClient().single_torrent_func(test_hash, other_param)
    FakeClient().single_torrent_func(test_hash, param=other_param)
    FakeClient().single_torrent_func(hash=test_hash, param=other_param)
    FakeClient().single_torrent_func(torrent_hash=test_hash, param=other_param)


@pytest.mark.parametrize(
    "test_hashes", (None, "", "xcvbxcvbxcvb", ("xcvbxcvbxcvb", "ertyertye"))
)
@pytest.mark.parametrize("other_param", (None, "", "irivgjkd"))
def test_hash_handler_multiple_hashes(test_hashes, other_param):
    class FakeClient:
        @handle_hashes
        def multiple_torrent_func(self, torrent_hashes=None, param=None):
            assert list2str(torrent_hashes) == list2str(test_hashes)
            assert param == other_param

    FakeClient().multiple_torrent_func(test_hashes, other_param)
    FakeClient().multiple_torrent_func(test_hashes, param=other_param)
    FakeClient().multiple_torrent_func(hashes=test_hashes, param=other_param)
    FakeClient().multiple_torrent_func(torrent_hashes=test_hashes, param=other_param)


def test_response_text():
    class ResponseTextTest:
        @response_text(str)
        def return_input_as_str(self, input):
            return input

        @response_text(dict)
        def return_input_as_dict(self, input):
            return input

    input = "asdf"
    assert input == ResponseTextTest().return_input_as_str(input)
    with pytest.raises(APIError):
        ResponseTextTest().return_input_as_dict(input)


def test_response_json():
    class MyDict(AttrDict):
        def __init__(self, *args):
            super(AttrDict, self).__init__(args[0])
            self._setattr("_allow_invalid_attributes", True)

    class FakeJson(AttrDict):
        data = None

        def __init__(self):
            super(AttrDict, self).__init__()
            self._setattr("_allow_invalid_attributes", True)

        @property
        def text(self):
            return dumps(self.data)

    class ResponseJsonTest:
        _SIMPLE_RESPONSES = False

        @response_json(dict)
        def return_input_as_dict(self, input):
            return input

        @response_json(MyDict)
        def return_input_as_mydict(self, input):
            return input

    input = {"one": 1}
    assert isinstance(ResponseJsonTest().return_input_as_dict(input), dict)
    input = FakeJson()
    input.data = {"one": 1}
    assert isinstance(ResponseJsonTest().return_input_as_mydict(input), MyDict)
    input.data = ["asdf"]
    with pytest.raises(APIError):
        ResponseJsonTest().return_input_as_mydict(input)


def test_version_implemented():
    class FakeClient:
        _RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS = True
        version = "1.0"

        def app_web_api_version(self):
            return self.version

        @endpoint_introduced("1.1", "test1")
        def endpoint_not_implemented(self):
            return

        @endpoint_introduced("0.9", "test2")
        def endpoint_implemented(self):
            return

    with pytest.raises(NotImplementedError):
        FakeClient().endpoint_not_implemented()

    assert FakeClient().endpoint_implemented() is None

    fake_client = FakeClient()
    fake_client._RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS = False
    assert fake_client.endpoint_not_implemented() is None


def test_version_removed():
    class FakeClient:
        _RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS = True
        version = "1.0"

        def app_web_api_version(self):
            return self.version

        @version_removed("0.0.0", "test1")
        def endpoint_not_implemented(self):
            return

        @version_removed("9999999", "test2")
        def endpoint_implemented(self):
            return

    with pytest.raises(NotImplementedError):
        FakeClient().endpoint_not_implemented()

    assert FakeClient().endpoint_implemented() is None

    fake_client = FakeClient()
    fake_client._RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS = False
    assert fake_client.endpoint_not_implemented() is None



from os import environ
from time import sleep

import pytest

from qbittorrentapi import APIConnectionError


def test_shutdown(client):
    if environ.get("CI") == "true":
        client.app.shutdown()
        with pytest.raises(APIConnectionError):
            for _ in range(100):
                client.app_version()
                sleep(0.1)



from os import environ
from os import path
from sys import path as sys_path
from time import sleep

import pytest
import six
from pkg_resources import parse_version as v

from qbittorrentapi import APIConnectionError
from qbittorrentapi import Client
from qbittorrentapi._version_support import (
    APP_VERSION_2_API_VERSION_MAP as api_version_map,
)

environ.setdefault("QBITTORRENTAPI_HOST", "localhost:8080")
environ.setdefault("QBITTORRENTAPI_USERNAME", "admin")
environ.setdefault("QBITTORRENTAPI_PASSWORD", "adminadmin")
try:
    environ.setdefault("QBT_VER", Client().app.version)
except APIConnectionError:
    print("is qBittorrent running???")

qbt_version = environ.get("QBT_VER", "")
qbt_version = qbt_version if qbt_version.startswith("v") else "v" + qbt_version

environ.setdefault("IS_QBT_DEV", "" if qbt_version in api_version_map else "1")
IS_QBT_DEV = bool(environ.get("IS_QBT_DEV", False))

BASE_PATH = sys_path[0]
RESOURCES = path.join(BASE_PATH, "tests", "resources")
assert BASE_PATH.split("/")[-1] == "qbittorrent-api"

_check_limit = 10

_orig_torrent_url = (
    "https://releases.ubuntu.com/22.04.1/ubuntu-22.04.1-desktop-amd64.iso.torrent"
)
_orig_torrent_hash = "3b245504cf5f11bbdbe1201cea6a6bf45aee1bc0"

torrent1_filename = "kubuntu-22.04.1-desktop-amd64.iso.torrent"
with open(path.join(RESOURCES, torrent1_filename), mode="rb") as f:
    torrent1_file = f.read()
torrent1_url = (
    "https://cdimage.ubuntu.com/kubuntu/releases/22.04/release/" + torrent1_filename
)
torrent1_hash = "8e905ea8e925c3a7fda9eb2e964a0e5c97cc2ce0"

torrent2_filename = "xubuntu-22.04.1-desktop-amd64.iso.torrent"
torrent2_url = (
    "https://cdimage.ubuntu.com/xubuntu/releases/22.04/release/" + torrent2_filename
)
torrent2_hash = "b813f485c0e6d17f6877c8d6942b3bdc7c227176"

with open(path.join(RESOURCES, "root_folder.torrent"), mode="rb") as f:
    root_folder_torrent_file = f.read()
root_folder_torrent_hash = "a14553bd936a6d496402082454a70ea7a9521adc"


def get_func(obj, method_name):
    for attr in method_name.split("."):
        obj = getattr(obj, attr)
    return obj


def mkpath(*user_path):
    if any(user_path):
        return path.abspath(path.realpath(path.expanduser(path.join(*user_path))))
    return ""


def check(
    check_func, value, reverse=False, negate=False, any=False, check_limit=_check_limit
):
    """
    Compare the return value of an arbitrary function to expected value with
    retries. Since some requests take some time to take effect in qBittorrent,
    the retries every second for 10 seconds.

    :param check_func: callable to generate values to check
    :param value: str, int, or iterator of values to look for
    :param reverse: False: look for check_func return in value; True: look for value in check_func return
    :param negate: False: value must be found; True: value must not be found
    :param check_limit: seconds to spend checking
    :param any: False: all values must be (not) found; True: any value must be (not) found
    """

    def _do_check(_check_func_val, _v, _negate, _reverse):
        if _negate:
            if _reverse:
                assert _v not in _check_func_val
            else:
                assert _check_func_val not in (_v,)
        else:
            if _reverse:
                assert _v in _check_func_val
            else:
                assert _check_func_val in (_v,)

    if isinstance(value, (six.string_types, int)):
        value = (value,)

    try:
        for i in range(_check_limit):
            try:
                exp = None
                for val in value:
                    # clear any previous exceptions if any=True
                    exp = None if any else exp

                    try:
                        # get val here so pytest includes value in failures
                        check_val = check_func()
                        _do_check(check_val, val, negate, reverse)
                    except AssertionError as e:
                        exp = e

                    # fail the test on first failure if any=False
                    if not any and exp:
                        break
                    # this value passed so test succeeded if any=True
                    if any and not exp:
                        break

                # raise caught inner exception for handling
                if exp:
                    raise exp

                # test succeeded!!!!
                break

            except AssertionError:
                if i >= check_limit - 1:
                    raise
                sleep(1)
    except APIConnectionError:
        raise AssertionError("qBittrorrent crashed...")


def retry(retries=3):
    """decorator to retry a function if there's an exception."""

    def inner(f):
        def wrapper(*args, **kwargs):
            for retry_count in range(retries):
                try:
                    return f(*args, **kwargs)
                except Exception:
                    if retry_count >= (retries - 1):
                        raise

        return wrapper

    return inner


@pytest.fixture(autouse=True)
def abort_if_qbittorrent_crashes(client):
    """Abort tests if qbittorrent disappears during testing."""
    try:
        client.app_version()
        yield
    except APIConnectionError:
        pytest.exit("qBittorrent crashed :(")


@pytest.fixture(scope="session")
def client():
    """qBittorrent Client for testing session."""
    try:
        client = Client(
            RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
            VERBOSE_RESPONSE_LOGGING=True,
            VERIFY_WEBUI_CERTIFICATE=False,
        )
        client.auth_log_in()
        # add orig_torrent to qBittorrent
        client.torrents_add(urls=_orig_torrent_url, upload_limit=10, download_limit=10)
        client.app.preferences = dict(
            # enable RSS fetching
            rss_processing_enabled=True,
            # prevent banning IPs
            web_ui_max_auth_fail_count=1000,
            web_ui_ban_duration=1,
        )
        return client
    except APIConnectionError as e:
        pytest.exit("qBittorrent was not running when tests started: %s" % repr(e))


@pytest.fixture(scope="session")
def orig_torrent_hash():
    """Torrent hash for the Xubuntu torrent loaded for testing."""
    return _orig_torrent_hash


@pytest.fixture(scope="function")
def orig_torrent(client, orig_torrent_hash):
    """Torrent to remain in qBittorrent for entirety of session."""
    try:
        check(
            lambda: len(
                [t for t in client.torrents_info() if t.hash == orig_torrent_hash]
            ),
            value=1,
        )
        return [t for t in client.torrents_info() if t.hash == orig_torrent_hash][0]
    except APIConnectionError:
        pytest.exit('Failed to find "orig_torrent"')


@pytest.fixture
def new_torrent(client):
    """Torrent that is added on demand to qBittorrent and then removed."""
    yield next(new_torrent_standalone(client))


def new_torrent_standalone(client, torrent_hash=torrent1_hash, **kwargs):
    def add_test_torrent(client):
        for attempt in range(_check_limit):
            if kwargs:
                client.torrents.add(**kwargs)
            else:
                client.torrents.add(
                    torrent_files=torrent1_file,
                    save_path=path.expanduser("~/test_download/"),
                    category="test_category",
                    is_paused=True,
                    upload_limit=1024,
                    download_limit=2048,
                    is_sequential_download=True,
                    is_first_last_piece_priority=True,
                )
            try:
                # not all versions of torrents_info() support passing a hash
                return [t for t in client.torrents_info() if t.hash == torrent_hash][0]
            except Exception:
                if attempt >= _check_limit - 1:
                    raise
                sleep(1)

    @retry
    def delete_test_torrent(torrent):
        torrent.delete(delete_files=True)
        check(
            lambda: [t.hash for t in torrent._client.torrents_info()],
            torrent.hash,
            reverse=True,
            negate=True,
        )

    try:
        torrent = add_test_torrent(client)
    except APIConnectionError:
        yield None  # if qBittorrent crashed, it'll get caught in abort fixture
    else:
        yield torrent
        try:
            delete_test_torrent(torrent)
        except APIConnectionError:
            pass  # if qBittorrent crashed, it'll get caught in abort fixture


@pytest.fixture(scope="session")
def app_version(client):
    """qBittorrent App Version being used for testing."""
    if IS_QBT_DEV:
        return client.app.version
    return qbt_version


@pytest.fixture(scope="session")
def api_version(client):
    """qBittorrent App API Version being used for testing."""
    try:
        return api_version_map[qbt_version]
    except KeyError as exp:
        if IS_QBT_DEV:
            return client.app.web_api_version
        raise exp


@pytest.fixture(scope="function")
def rss_feed(client, api_version):
    def delete_feed(name):
        try:
            client.rss_remove_item(item_path=name)
            check(lambda: client.rss_items(), name, reverse=True, negate=True)
        except Exception:
            pass

    if v(api_version) >= v("2.2"):
        name = "DistroWatch"
        url = "https://distrowatch.com/news/torrents.xml"
        client.app.preferences = dict(rss_auto_downloading_enabled=False)
        # refreshing the feed is finicky...so try several times if necessary
        done = False
        for i in range(30):
            delete_feed(name)
            client.rss.add_feed(url=url, item_path=name)
            # refreshing too quickly can cause it to never refresh.
            # auto-update is enabled in preferences
            # client.rss.refresh_item(item_path=name)  # noqa: E800
            # wait until the rss feed exists and is refreshed
            check(lambda: client.rss_items(), name, reverse=True)
            # wait until feed is refreshed
            for j in range(10):
                if client.rss.items.with_data[name]["articles"]:
                    done = True
                    yield name
                    delete_feed(name)
                    break
                sleep(0.5)
            if done:
                break
        else:
            raise Exception("RSS Feed '%s' did not refresh..." % name)
    else:
        yield ""


def pytest_sessionfinish(session, exitstatus):
    try:
        if environ.get("CI") != "true":
            client = Client()
            # remove all torrents
            for torrent in client.torrents_info():
                client.torrents_delete(delete_files=True, torrent_hashes=torrent.hash)
    except Exception:
        pass



import pytest
from pkg_resources import parse_version as v

from qbittorrentapi.transfer import TransferInfoDictionary


def test_info(client):
    info = client.transfer_info()
    assert isinstance(info, TransferInfoDictionary)
    assert "connection_status" in info
    info = client.transfer.info
    assert isinstance(info, TransferInfoDictionary)
    assert "connection_status" in info


def test_speed_limits_mode(client):
    assert client.transfer_speed_limits_mode() in {"0", "1"}
    assert client.transfer.speed_limits_mode in {"0", "1"}

    original_mode = client.transfer.speed_limits_mode
    client.transfer_toggle_speed_limits_mode()
    assert client.transfer.speed_limits_mode != original_mode
    original_mode = client.transfer.speed_limits_mode
    client.transfer.toggle_speed_limits_mode()
    assert client.transfer.speed_limits_mode != original_mode

    if client.transfer.speed_limits_mode == "0":
        client.transfer_toggle_speed_limits_mode(intended_state=True)
        assert client.transfer.speed_limits_mode != "0"
    if client.transfer.speed_limits_mode == "1":
        client.transfer_toggle_speed_limits_mode(intended_state=False)
        assert client.transfer.speed_limits_mode != "1"

    if client.transfer.speed_limits_mode == "0":
        client.transfer.speed_limits_mode = True
        assert client.transfer.speed_limits_mode != "0"
    if client.transfer.speed_limits_mode == "1":
        client.transfer.speed_limits_mode = False
        assert client.transfer.speed_limits_mode != "1"

    if client.transfer.speedLimitsMode == "0":
        client.transfer.speedLimitsMode = True
        assert client.transfer.speedLimitsMode != "0"
    if client.transfer.speedLimitsMode == "1":
        client.transfer.speedLimitsMode = False
        assert client.transfer.speedLimitsMode != "1"

    client.transfer_toggle_speed_limits_mode(intended_state=False)
    assert client.transfer.speed_limits_mode != "1"


def test_download_limit(client):
    client.transfer_set_download_limit(limit=2048)
    assert client.transfer_download_limit() == 2048
    client.transfer_setDownloadLimit(limit=3072)
    assert client.transfer_downloadLimit() == 3072

    client.transfer.download_limit = 4096
    assert client.transfer.download_limit == 4096
    client.transfer.downloadLimit = 5120
    assert client.transfer.downloadLimit == 5120


def test_upload_limit(client):
    client.transfer_set_upload_limit(limit=2048)
    assert client.transfer_upload_limit() == 2048
    client.transfer_setUploadLimit(limit=3072)
    assert client.transfer_uploadLimit() == 3072

    client.transfer.upload_limit = 4096
    assert client.transfer.upload_limit == 4096
    client.transfer.uploadLimit = 5120
    assert client.transfer.uploadLimit == 5120


def test_ban_peers(client, api_version):
    if v(api_version) >= v("2.3"):
        client.transfer_ban_peers(peers="1.1.1.1:8080")
        assert "1.1.1.1" in client.app.preferences.banned_IPs
        client.transfer.ban_peers(peers="1.1.1.2:8080")
        assert "1.1.1.2" in client.app.preferences.banned_IPs

        client.transfer_ban_peers(peers=["1.1.1.3:8080", "1.1.1.4:8080"])
        assert "1.1.1.3" in client.app.preferences.banned_IPs
        assert "1.1.1.4" in client.app.preferences.banned_IPs
        client.transfer.ban_peers(peers=["1.1.1.5:8080", "1.1.1.6:8080"])
        assert "1.1.1.5" in client.app.preferences.banned_IPs
        assert "1.1.1.6" in client.app.preferences.banned_IPs
    else:
        with pytest.raises(NotImplementedError):
            client.transfer_ban_peers(peers="1.1.1.1:8080")



import pytest

from qbittorrentapi.log import LogMainList
from qbittorrentapi.log import LogPeersList


@pytest.mark.parametrize("normal", (True, False))
@pytest.mark.parametrize("info", (True, False))
@pytest.mark.parametrize("warning", (True, False))
@pytest.mark.parametrize("critical", (True, False))
@pytest.mark.parametrize("last_known_id", (None, 0, 100000))
def test_main(client, normal, info, warning, critical, last_known_id):
    assert isinstance(
        client.log_main(
            normal=normal,
            info=info,
            warning=warning,
            critical=critical,
            last_known_id=last_known_id,
        ),
        LogMainList,
    )
    assert isinstance(
        client.log.main(
            normal=normal,
            info=info,
            warning=warning,
            critical=critical,
            last_known_id=last_known_id,
        ),
        LogMainList,
    )
    assert isinstance(client.log.main.info(last_known_id=last_known_id), LogMainList)
    assert isinstance(client.log.main.normal(last_known_id=last_known_id), LogMainList)
    assert isinstance(client.log.main.warning(last_known_id=last_known_id), LogMainList)
    assert isinstance(
        client.log.main.critical(last_known_id=last_known_id), LogMainList
    )


@pytest.mark.parametrize("last_known_id", (None, 0, 100000))
def test_log_peers(client, last_known_id):
    assert isinstance(client.log_peers(last_known_id=last_known_id), LogPeersList)
    assert isinstance(client.log.peers(last_known_id=last_known_id), LogPeersList)



import errno
import logging
import platform
from sys import version_info
from time import sleep

from pkg_resources import parse_version as v

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable  # noqa: F401

import pytest
import requests

from qbittorrentapi.exceptions import Conflict409Error
from qbittorrentapi.exceptions import Forbidden403Error
from qbittorrentapi.exceptions import InvalidRequest400Error
from qbittorrentapi.exceptions import TorrentFileError
from qbittorrentapi.exceptions import TorrentFileNotFoundError
from qbittorrentapi.exceptions import TorrentFilePermissionError
from qbittorrentapi.torrents import TagList
from qbittorrentapi.torrents import TorrentCategoriesDictionary
from qbittorrentapi.torrents import TorrentFilesList
from qbittorrentapi.torrents import TorrentInfoList
from qbittorrentapi.torrents import TorrentLimitsDictionary
from qbittorrentapi.torrents import TorrentPieceInfoList
from qbittorrentapi.torrents import TorrentPropertiesDictionary
from qbittorrentapi.torrents import TorrentsAddPeersDictionary
from qbittorrentapi.torrents import TrackersList
from qbittorrentapi.torrents import WebSeedsList
from tests.conftest import check
from tests.conftest import get_func
from tests.conftest import mkpath
from tests.conftest import new_torrent_standalone
from tests.conftest import retry
from tests.conftest import root_folder_torrent_file
from tests.conftest import root_folder_torrent_hash
from tests.conftest import torrent1_filename
from tests.conftest import torrent1_hash
from tests.conftest import torrent1_url
from tests.conftest import torrent2_filename
from tests.conftest import torrent2_hash
from tests.conftest import torrent2_url


def disable_queueing(client):
    if client.app.preferences.queueing_enabled:
        client.app.set_preferences(dict(queueing_enabled=False))


def enable_queueing(client):
    if not client.app.preferences.queueing_enabled:
        client.app.set_preferences(dict(queueing_enabled=True))


@pytest.mark.parametrize(
    "client_func",
    (("torrents_add", "torrents_delete"), ("torrents.add", "torrents.delete")),
)
def test_add_delete(client, api_version, client_func):
    def download_file(url, filename=None, return_bytes=False):
        max_attempts = 3
        for attempt in range(max_attempts):
            try:
                with requests.get(url) as r:
                    r.raise_for_status()
                    if return_bytes:
                        return r.content
                    with open(mkpath("~/%s" % filename), "wb") as f:
                        for chunk in r.iter_content(chunk_size=1024):
                            f.write(chunk)
            except (Exception if attempt < (max_attempts - 1) else ZeroDivisionError):
                pass  # throw away errors until we hit the retry limit
            else:
                return
        raise Exception("Download failed: %s" % url)

    def delete():
        get_func(client, client_func[1])(
            delete_files=True, torrent_hashes=torrent1_hash
        )
        get_func(client, client_func[1])(
            delete_files=True, torrent_hashes=torrent2_hash
        )
        check(
            lambda: [t.hash for t in client.torrents_info()],
            torrent2_hash,
            reverse=True,
            negate=True,
        )

    def check_torrents_added(f):
        def inner(**kwargs):
            try:
                f(**kwargs)
                check(
                    lambda: [t.hash for t in client.torrents_info()],
                    torrent1_hash,
                    reverse=True,
                )
                if kwargs.get("single", False) is False:
                    check(
                        lambda: [t.hash for t in client.torrents_info()],
                        torrent2_hash,
                        reverse=True,
                    )
            finally:
                sleep(1)
                delete()

        return inner

    @check_torrents_added
    def add_by_filename(single):
        download_file(url=torrent1_url, filename=torrent1_filename)
        download_file(url=torrent2_url, filename=torrent2_filename)
        files = ("~/%s" % torrent1_filename, "~/%s" % torrent2_filename)

        if single:
            assert get_func(client, client_func[0])(torrent_files=files[0]) == "Ok."
        else:
            assert get_func(client, client_func[0])(torrent_files=files) == "Ok."

    @check_torrents_added
    def add_by_filename_dict(single):
        download_file(url=torrent1_url, filename=torrent1_filename)
        download_file(url=torrent2_url, filename=torrent2_filename)

        if single:
            assert (
                get_func(client, client_func[0])(
                    torrent_files={torrent1_filename: "~/%s" % torrent1_filename}
                )
                == "Ok."
            )
        else:
            files = {
                torrent1_filename: "~/%s" % torrent1_filename,
                torrent2_filename: "~/%s" % torrent2_filename,
            }
            assert get_func(client, client_func[0])(torrent_files=files) == "Ok."

    @check_torrents_added
    def add_by_filehandles(single):
        download_file(url=torrent1_url, filename=torrent1_filename)
        download_file(url=torrent2_url, filename=torrent2_filename)
        files = (
            open(mkpath("~/" + torrent1_filename), "rb"),
            open(mkpath("~/" + torrent2_filename), "rb"),
        )

        if single:
            assert get_func(client, client_func[0])(torrent_files=files[0]) == "Ok."
        else:
            assert get_func(client, client_func[0])(torrent_files=files) == "Ok."

        for file in files:
            file.close()

    @check_torrents_added
    def add_by_bytes(single):
        files = (
            download_file(torrent1_url, return_bytes=True),
            download_file(torrent2_url, return_bytes=True),
        )

        if single:
            assert get_func(client, client_func[0])(torrent_files=files[0]) == "Ok."
        else:
            assert get_func(client, client_func[0])(torrent_files=files) == "Ok."

    @check_torrents_added
    def add_by_url(single):
        urls = (torrent1_url, torrent2_url)

        if single:
            get_func(client, client_func[0])(urls=urls[0])
        else:
            get_func(client, client_func[0])(urls=urls)

    if v(api_version) >= v("2.0.1"):
        # something was wrong with torrents_add on v2.0.0 (the initial version)
        add_by_filename(single=False)
        add_by_filename(single=True)
        add_by_filename_dict(single=False)
        add_by_filename_dict(single=True)
        add_by_url(single=False)
        add_by_url(single=True)
        add_by_filehandles(single=False)
        add_by_filehandles(single=True)
        add_by_bytes(single=False)
        add_by_bytes(single=True)


def test_add_torrent_file_fail(client, monkeypatch):
    # torrent add is wonky in python2 because of support for raw bytes...
    if version_info[0] > 2:
        with pytest.raises(TorrentFileNotFoundError):
            client.torrents_add(torrent_files="/tmp/asdfasdfasdfasdf")

        with pytest.raises(TorrentFilePermissionError):
            client.torrents_add(torrent_files="/etc/shadow")

        if platform.python_implementation() == "CPython":
            with pytest.raises(TorrentFileError):

                def fake_open(*arg, **kwargs):
                    raise IOError(errno.ENODEV)

                with monkeypatch.context() as m:
                    m.setitem(__builtins__, "open", fake_open)
                    client.torrents_add(torrent_files="/etc/hosts")


def test_close_file_fail(client, monkeypatch, caplog):
    def fake_norm_files(files):
        return {object: object}, (object,)

    def post(*args, **kwargs):
        return "OK"

    if version_info[0] > 2:
        with monkeypatch.context() as m:
            m.setattr(client, "_normalize_torrent_files", fake_norm_files)
            m.setattr(client, "_post", post)
            with caplog.at_level(logging.WARNING, logger="qbittorrentapi"):
                client.torrents_add(torrent_files=object)
                assert "Failed to close file" in caplog.text


@pytest.mark.parametrize("keep_root_folder", (True, False, None))
@pytest.mark.parametrize(
    "content_layout", (None, "Original", "Subfolder", "NoSubfolder")
)
def test_add_options(client, api_version, keep_root_folder, content_layout):
    client.torrents_delete(torrent_hashes=root_folder_torrent_hash, delete_files=True)
    if v(api_version) >= v("2.3.0"):
        client.torrents_create_tags("option-tag")
    torrent = next(
        new_torrent_standalone(
            torrent_hash=root_folder_torrent_hash,
            client=client,
            torrent_files=root_folder_torrent_file,
            save_path=mkpath("~/test_download/"),
            category="test_category",
            is_paused=True,
            upload_limit=1024,
            download_limit=2048,
            is_sequential_download=True,
            is_first_last_piece_priority=True,
            is_root_folder=keep_root_folder,
            rename="this is a new name for the torrent",
            use_auto_torrent_management=False,
            tags="option-tag",
            content_layout=content_layout,
            ratio_limit=2,
            seeding_time_limit=120,
        )
    )
    check(lambda: torrent.info.category, "test_category")
    check(
        lambda: torrent.info.state,
        ("pausedDL", "checkingResumeData"),
        reverse=True,
        any=True,
    )
    check(lambda: mkpath(torrent.info.save_path), mkpath("~/test_download/"))
    check(lambda: torrent.info.up_limit, 1024)
    check(lambda: torrent.info.dl_limit, 2048)
    check(lambda: torrent.info.seq_dl, True)
    if v(api_version) >= v("2.0.1"):
        check(lambda: torrent.info.f_l_piece_prio, True)
    if content_layout is None:
        check(
            lambda: torrent.files[0]["name"].startswith("root_folder"),
            keep_root_folder in {True, None},
        )
    check(lambda: torrent.info.name, "this is a new name for the torrent")
    check(lambda: torrent.info.auto_tmm, False)
    if v(api_version) >= v("2.6.2"):
        check(lambda: torrent.info.tags, "option-tag")

    if v(api_version) >= v("2.7"):
        # after web api v2.7...root dir is driven by content_layout
        if content_layout is None:
            should_root_dir_exists = keep_root_folder in {None, True}
        else:
            should_root_dir_exists = content_layout in {"Original", "Subfolder"}
    else:
        # before web api v2.7...it is driven by is_root_folder
        if content_layout is not None and keep_root_folder is None:
            should_root_dir_exists = content_layout in {"Original", "Subfolder"}
        else:
            should_root_dir_exists = keep_root_folder in {None, True}
    check(
        lambda: any(f["name"].startswith("root_folder") for f in torrent.files),
        should_root_dir_exists,
    )

    if v(api_version) >= v("2.8.1"):
        check(lambda: torrent.info.ratio_limit, 2)
        check(lambda: torrent.info.seeding_time_limit, 120)


@pytest.mark.parametrize("use_download_path", (None, True, False))
def test_torrents_add_download_path(client, api_version, use_download_path):
    if v(api_version) >= v("2.8.4"):
        client.torrents_delete(
            torrent_hashes=root_folder_torrent_hash, delete_files=True
        )
        save_path = mkpath("/tmp", "down_path_save_path_test")
        download_path = mkpath("/tmp", "down_path_test")
        torrent = next(
            new_torrent_standalone(
                client=client,
                torrent_hash=root_folder_torrent_hash,
                torrent_files=root_folder_torrent_file,
                download_path=download_path,
                use_download_path=use_download_path,
                test_download_limit=1024,
                save_path=save_path,
            )
        )

        if use_download_path is False:
            check(
                lambda: mkpath(torrent.info.download_path), download_path, negate=True
            )
        else:
            check(lambda: mkpath(torrent.info.download_path), download_path)


def test_properties(client, orig_torrent):
    props = client.torrents_properties(torrent_hash=orig_torrent.hash)
    assert isinstance(props, TorrentPropertiesDictionary)


def test_trackers(client, orig_torrent):
    trackers = client.torrents_trackers(torrent_hash=orig_torrent.hash)
    assert isinstance(trackers, TrackersList)


def test_webseeds(client, orig_torrent):
    web_seeds = client.torrents_webseeds(torrent_hash=orig_torrent.hash)
    assert isinstance(web_seeds, WebSeedsList)


def test_files(client, orig_torrent):
    files = client.torrents_files(torrent_hash=orig_torrent.hash)
    assert isinstance(files, TorrentFilesList)
    assert "availability" in files[0]

    assert all(file["id"] == file["index"] for file in files)


@pytest.mark.parametrize(
    "client_func", ("torrents_piece_states", "torrents_pieceStates")
)
def test_piece_states(client, orig_torrent, client_func):
    piece_states = get_func(client, client_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_states, TorrentPieceInfoList)


@pytest.mark.parametrize(
    "client_func", ("torrents_piece_hashes", "torrents_pieceHashes")
)
def test_piece_hashes(client, orig_torrent, client_func):
    piece_hashes = get_func(client, client_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_hashes, TorrentPieceInfoList)


@pytest.mark.parametrize("trackers", ("127.0.0.1", ("127.0.0.2", "127.0.0.3")))
@pytest.mark.parametrize(
    "client_func", ("torrents_add_trackers", "torrents_addTrackers")
)
def test_add_trackers(client, trackers, client_func, new_torrent):
    get_func(client, client_func)(torrent_hash=new_torrent.hash, urls=trackers)
    check(lambda: (t.url for t in new_torrent.trackers), trackers, reverse=True)


@pytest.mark.parametrize(
    "client_func", ("torrents_edit_tracker", "torrents_editTracker")
)
def test_edit_tracker(client, api_version, client_func, orig_torrent):
    if v(api_version) >= v("2.2.0"):
        orig_torrent.add_trackers("127.1.0.1")
        get_func(client, client_func)(
            torrent_hash=orig_torrent.hash,
            original_url="127.1.0.1",
            new_url="127.1.0.2",
        )
        check(lambda: (t.url for t in orig_torrent.trackers), "127.1.0.2", reverse=True)
        get_func(client, "torrents_remove_trackers")(
            torrent_hash=orig_torrent.hash, urls="127.1.0.2"
        )
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(
                torrent_hash=orig_torrent.hash,
                original_url="127.1.0.1",
                new_url="127.1.0.2",
            )


@pytest.mark.parametrize(
    "trackers",
    (
        ("127.2.0.1",),
        ("127.2.0.2", "127.2.0.3"),
    ),
)
@pytest.mark.parametrize(
    "client_func", ("torrents_remove_trackers", "torrents_removeTrackers")
)
def test_remove_trackers(client, api_version, trackers, client_func, orig_torrent):
    if v(api_version) >= v("2.2.0"):
        orig_torrent.add_trackers(trackers)
        get_func(client, client_func)(torrent_hash=orig_torrent.hash, urls=trackers)
        check(
            lambda: (t.url for t in orig_torrent.trackers),
            trackers,
            reverse=True,
            negate=True,
        )
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(torrent_hash=orig_torrent.hash, urls=trackers)


@pytest.mark.parametrize("client_func", ("torrents_file_priority", "torrents_filePrio"))
def test_file_priority(client, orig_torrent, orig_torrent_hash, client_func):
    get_func(client, client_func)(
        torrent_hash=orig_torrent_hash, file_ids=0, priority=6
    )
    check(lambda: orig_torrent.files[0].priority, 6)
    get_func(client, client_func)(
        torrent_hash=orig_torrent_hash, file_ids=0, priority=7
    )
    check(lambda: orig_torrent.files[0].priority, 7)


@pytest.mark.parametrize("new_name", (("new name 2", "new_name_2"),))
def test_rename(client, orig_torrent_hash, orig_torrent, new_name):
    client.torrents_rename(torrent_hash=orig_torrent_hash, new_torrent_name=new_name[0])
    check(lambda: orig_torrent.info.name.replace("+", " "), new_name[0])


@pytest.mark.parametrize("new_name", ("new name file 2", "new_name_file_2"))
@pytest.mark.parametrize("client_func", ("torrents_rename_file", "torrents_renameFile"))
def test_rename_file(
    client,
    api_version,
    new_torrent,
    new_name,
    client_func,
):
    if v(api_version) >= v("2.4.0"):
        sleep(1)
        # pre-v4.3.3 rename_file signature
        get_func(client, client_func)(
            torrent_hash=new_torrent.hash, file_id=0, new_file_name=new_name
        )
        check(lambda: new_torrent.files[0].name.replace("+", " "), new_name)
        # test invalid file ID is rejected
        with pytest.raises(Conflict409Error):
            get_func(client, client_func)(
                torrent_hash=new_torrent.hash, file_id=10, new_file_name=new_name
            )
        # post-v4.3.3 rename_file signature
        new_new_name = new_name + "NEW"
        get_func(client, client_func)(
            torrent_hash=new_torrent.hash,
            old_path=new_torrent.files[0].name,
            new_path=new_new_name,
        )
        check(lambda: new_torrent.files[0].name.replace("+", " "), new_new_name)
        # test invalid old_path is rejected
        with pytest.raises(Conflict409Error):
            get_func(client, client_func)(
                torrent_hash=new_torrent.hash, old_path="asdf", new_path="xcvb"
            )
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(
                torrent_hash=new_torrent.hash, file_id=0, new_file_name=new_name
            )


@pytest.mark.parametrize("new_name", ("asdf zxcv", "asdf_zxcv"))
@pytest.mark.parametrize(
    "client_func", ("torrents_rename_folder", "torrents_renameFolder")
)
def test_rename_folder(client, app_version, new_torrent, new_name, client_func):
    if v(app_version) >= v("v4.3.3"):
        # move the file in to a new folder
        orig_file_path = new_torrent.files[0].name
        new_folder = "qwer"
        client.torrents_rename_file(
            torrent_hash=new_torrent.hash,
            old_path=orig_file_path,
            new_path=new_folder + "/" + orig_file_path,
        )
        sleep(1)  # qBittorrent crashes if you make these calls too fast...
        # test rename that new folder
        get_func(client, client_func)(
            torrent_hash=new_torrent.hash,
            old_path=new_folder,
            new_path=new_name,
        )
        check(
            lambda: new_torrent.files[0].name.replace("+", " "),
            new_name + "/" + orig_file_path,
        )
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(
                torrent_hash="asdf", old_path="asdf", new_path="zxcv"
            )


@pytest.mark.parametrize("client_func", ("torrents_info", "torrents.info"))
def test_torrents_info(client, orig_torrent_hash, client_func):
    assert isinstance(get_func(client, client_func)(), TorrentInfoList)
    if "." in client_func:
        assert isinstance(get_func(client, client_func).all(), TorrentInfoList)
        assert isinstance(get_func(client, client_func).downloading(), TorrentInfoList)
        assert isinstance(get_func(client, client_func).completed(), TorrentInfoList)
        assert isinstance(get_func(client, client_func).paused(), TorrentInfoList)
        assert isinstance(get_func(client, client_func).active(), TorrentInfoList)
        assert isinstance(get_func(client, client_func).inactive(), TorrentInfoList)
        assert isinstance(get_func(client, client_func).resumed(), TorrentInfoList)
        assert isinstance(get_func(client, client_func).stalled(), TorrentInfoList)
        assert isinstance(
            get_func(client, client_func).stalled_uploading(), TorrentInfoList
        )
        assert isinstance(
            get_func(client, client_func).stalled_downloading(), TorrentInfoList
        )


@pytest.mark.parametrize("client_func", ("torrents_info", "torrents.info"))
def test_torrents_info_tag(client, api_version, new_torrent, client_func):
    if v(api_version) >= v("2.8.3"):
        tag_name = "tag_filter_name"
        client.torrents_add_tags(tags=tag_name, torrent_hashes=new_torrent.hash)
        torrents = get_func(client, client_func)(
            torrent_hashes=new_torrent.hash, tag=tag_name
        )
        assert new_torrent.hash in {t.hash for t in torrents}


@pytest.mark.parametrize(
    "client_func",
    (("torrents_pause", "torrents_resume"), ("torrents.pause", "torrents.resume")),
)
def test_pause_resume(client, orig_torrent, orig_torrent_hash, client_func):
    get_func(client, client_func[0])(torrent_hashes=orig_torrent_hash)
    check(
        lambda: client.torrents_info(torrents_hashes=orig_torrent.hash)[0].state,
        ("stalledDL", "pausedDL"),
        any=True,
        check_limit=20,
    )

    get_func(client, client_func[1])(torrent_hashes=orig_torrent_hash)
    check(
        lambda: client.torrents_info(torrents_hashes=orig_torrent.hash)[0].state,
        "pausedDL",
        negate=True,
        check_limit=20,
    )


def test_action_for_all_torrents(client):
    client.torrents.resume.all()
    for torrent in client.torrents.info():
        check(
            lambda: client.torrents_info(torrents_hashes=torrent.hash)[0].state,
            ("pausedDL",),
            negate=True,
        )
    client.torrents.pause.all()
    for torrent in client.torrents.info():
        check(
            lambda: client.torrents_info(torrents_hashes=torrent.hash)[0].state,
            ("stalledDL", "pausedDL"),
            any=True,
        )


@pytest.mark.parametrize("client_func", ("torrents_recheck", "torrents.recheck"))
def test_recheck(client, orig_torrent_hash, client_func):
    get_func(client, client_func)(torrent_hashes=orig_torrent_hash)


@pytest.mark.parametrize("client_func", ("torrents_reannounce", "torrents.reannounce"))
def test_reannounce(client, api_version, orig_torrent_hash, client_func):
    if v(api_version) >= v("2.0.2"):
        get_func(client, client_func)(torrent_hashes=orig_torrent_hash)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(torrent_hashes=orig_torrent_hash)


@pytest.mark.parametrize(
    "client_func",
    (
        (
            "torrents_increase_priority",
            "torrents_decrease_priority",
            "torrents_top_priority",
            "torrents_bottom_priority",
        ),
        (
            "torrents_increasePrio",
            "torrents_decreasePrio",
            "torrents_topPrio",
            "torrents_bottomPrio",
        ),
    ),
)
def test_priority(client, new_torrent, client_func):
    disable_queueing(client)

    with pytest.raises(Conflict409Error):
        get_func(client, client_func[0])(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        get_func(client, client_func[1])(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        get_func(client, client_func[2])(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        get_func(client, client_func[3])(torrent_hashes=new_torrent.hash)

    enable_queueing(client)

    @retry
    def test1(current_priority):
        get_func(client, client_func[0])(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority < current_priority, True)

    @retry
    def test2(current_priority):
        get_func(client, client_func[1])(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority > current_priority, True)

    @retry
    def test3(current_priority):
        get_func(client, client_func[2])(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority < current_priority, True)

    @retry
    def test4(current_priority):
        get_func(client, client_func[3])(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority > current_priority, True)

    current_priority = new_torrent.info.priority
    test1(current_priority)
    current_priority = new_torrent.info.priority
    test2(current_priority)
    current_priority = new_torrent.info.priority
    test3(current_priority)
    current_priority = new_torrent.info.priority
    test4(current_priority)


@pytest.mark.parametrize(
    "client_func",
    (
        ("torrents_set_download_limit", "torrents_download_limit"),
        ("torrents_setDownloadLimit", "torrents_downloadLimit"),
        ("torrents.set_download_limit", "torrents.download_limit"),
        ("torrents.setDownloadLimit", "torrents.downloadLimit"),
    ),
)
def test_download_limit(client, client_func, orig_torrent):
    orig_download_limit = get_func(client, client_func[1])(
        torrent_hashes=orig_torrent.hash
    )[orig_torrent.hash]

    get_func(client, client_func[0])(torrent_hashes=orig_torrent.hash, limit=100)
    assert isinstance(
        get_func(client, client_func[1])(torrent_hashes=orig_torrent.hash),
        TorrentLimitsDictionary,
    )
    check(
        lambda: get_func(client, client_func[1])(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        100,
    )

    # reset download limit
    get_func(client, client_func[0])(
        torrent_hashes=orig_torrent.hash, limit=orig_download_limit
    )
    check(
        lambda: get_func(client, client_func[1])(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        orig_download_limit,
    )


@pytest.mark.parametrize(
    "client_func",
    (
        ("torrents_set_upload_limit", "torrents_upload_limit"),
        ("torrents_setUploadLimit", "torrents_uploadLimit"),
        ("torrents.set_upload_limit", "torrents.upload_limit"),
        ("torrents.setUploadLimit", "torrents.uploadLimit"),
    ),
)
def test_upload_limit(client, client_func, orig_torrent):
    orig_upload_limit = get_func(client, client_func[1])(
        torrent_hashes=orig_torrent.hash
    )[orig_torrent.hash]

    get_func(client, client_func[0])(torrent_hashes=orig_torrent.hash, limit=100)
    assert isinstance(
        get_func(client, client_func[1])(torrent_hashes=orig_torrent.hash),
        TorrentLimitsDictionary,
    )
    check(
        lambda: get_func(client, client_func[1])(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        100,
    )

    # reset upload limit
    get_func(client, client_func[0])(
        torrent_hashes=orig_torrent.hash, limit=orig_upload_limit
    )
    check(
        lambda: get_func(client, client_func[1])(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        orig_upload_limit,
    )


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_set_share_limits",
        "torrents_setShareLimits",
        "torrents.set_share_limits",
        "torrents.setShareLimits",
    ),
)
def test_set_share_limits(client, api_version, client_func, orig_torrent):
    if v(api_version) >= v("2.0.1"):
        get_func(client, client_func)(
            ratio_limit=2, seeding_time_limit=5, torrent_hashes=orig_torrent.hash
        )
        check(lambda: orig_torrent.info.max_ratio, 2)
        check(lambda: orig_torrent.info.max_seeding_time, 5)
        get_func(client, client_func)(
            ratio_limit=3, seeding_time_limit=6, torrent_hashes=orig_torrent.hash
        )
        check(lambda: orig_torrent.info.max_ratio, 3)
        check(lambda: orig_torrent.info.max_seeding_time, 6)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(
                ratio_limit=2, seeding_time_limit=5, torrent_hashes=orig_torrent.hash
            )


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_set_location",
        "torrents_setLocation",
        "torrents.set_location",
        "torrents.setLocation",
    ),
)
def test_set_location(client, api_version, client_func, new_torrent):
    if v(api_version) >= v("2.0.2"):
        with pytest.raises(Forbidden403Error):
            get_func(client, client_func)(
                location="/etc/", torrent_hashes=new_torrent.hash
            )

        loc = mkpath("/tmp", "1")
        get_func(client, client_func)(location=loc, torrent_hashes=new_torrent.hash)
        # qBittorrent may return trailing separators depending on version....
        check(lambda: mkpath(new_torrent.info.save_path), loc, any=True)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_set_save_path",
        "torrents_setSavePath",
        "torrents.set_save_path",
        "torrents.setSavePath",
    ),
)
def test_set_save_path(client, api_version, client_func, new_torrent):
    if v(api_version) >= v("2.8.4"):
        with pytest.raises(Forbidden403Error):
            get_func(client, client_func)(
                save_path="/etc/", torrent_hashes=new_torrent.hash
            )
        with pytest.raises(Conflict409Error):
            get_func(client, client_func)(
                save_path="/etc/asdf", torrent_hashes=new_torrent.hash
            )

        loc = mkpath("/tmp", "savepath1")
        get_func(client, client_func)(save_path=loc, torrent_hashes=new_torrent.hash)
        # qBittorrent may return trailing separators depending on version....
        check(lambda: mkpath(new_torrent.info.save_path), loc, any=True)

    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(
                save_path="/etc/", torrent_hashes=new_torrent.hash
            )


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_set_download_path",
        "torrents_setDownloadPath",
        "torrents.set_download_path",
        "torrents.setDownloadPath",
    ),
)
def test_set_download_path(client, api_version, client_func, new_torrent):
    if v(api_version) >= v("2.8.4"):
        with pytest.raises(Forbidden403Error):
            get_func(client, client_func)(
                download_path="/etc/", torrent_hashes=new_torrent.hash
            )
        with pytest.raises(Conflict409Error):
            get_func(client, client_func)(
                download_path="/etc/asdf", torrent_hashes=new_torrent.hash
            )

        loc = mkpath("/tmp", "savepath1")
        get_func(client, client_func)(
            download_path=loc, torrent_hashes=new_torrent.hash
        )
        # qBittorrent may return trailing separators depending on version....
        check(lambda: mkpath(new_torrent.info.download_path), loc, any=True)

    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(
                save_path="/etc/", torrent_hashes=new_torrent.hash
            )


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_set_category",
        "torrents_setCategory",
        "torrents.set_category",
        "torrents.setCategory",
    ),
)
@pytest.mark.parametrize("name", ("awesome cat", "awesome_cat"))
def test_set_category(client, client_func, name, orig_torrent):
    with pytest.raises(Conflict409Error):
        get_func(client, client_func)(
            category="/!@#$%^&*(", torrent_hashes=orig_torrent.hash
        )

    client.torrents_create_category(name=name)
    try:
        get_func(client, client_func)(category=name, torrent_hashes=orig_torrent.hash)
        check(lambda: orig_torrent.info.category.replace("+", " "), name)
    finally:
        client.torrents_remove_categories(categories=name)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_set_auto_management",
        "torrents_setAutoManagement",
        "torrents.set_auto_management",
        "torrents.setAutoManagement",
    ),
)
def test_torrents_set_auto_management(client, client_func, orig_torrent):
    current_setting = orig_torrent.info.auto_tmm
    get_func(client, client_func)(
        enable=(not current_setting), torrent_hashes=orig_torrent.hash
    )
    check(lambda: orig_torrent.info.auto_tmm, (not current_setting))
    get_func(client, client_func)(
        enable=False, torrent_hashes=orig_torrent.hash
    )  # leave on False


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_toggle_sequential_download",
        "torrents_toggleSequentialDownload",
        "torrents.toggle_sequential_download",
        "torrents.toggleSequentialDownload",
    ),
)
def test_toggle_sequential_download(client, client_func, orig_torrent):
    current_setting = orig_torrent.info.seq_dl
    get_func(client, client_func)(torrent_hashes=orig_torrent.hash)
    check(lambda: orig_torrent.info.seq_dl, not current_setting)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_toggle_first_last_piece_priority",
        "torrents_toggleFirstLastPiecePrio",
        "torrents.toggle_first_last_piece_priority",
        "torrents.toggleFirstLastPiecePrio",
    ),
)
def test_toggle_first_last_piece_priority(
    client, api_version, client_func, new_torrent
):
    if v(api_version) >= v("2.0.1"):
        current_setting = new_torrent.info.f_l_piece_prio
        sleep(1)
        get_func(client, client_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.f_l_piece_prio, not current_setting)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_set_force_start",
        "torrents_setForceStart",
        "torrents.set_force_start",
        "torrents.setForceStart",
    ),
)
def test_set_force_start(client, client_func, orig_torrent):
    current_setting = orig_torrent.info.force_start
    get_func(client, client_func)(
        enable=(not current_setting), torrent_hashes=orig_torrent.hash
    )
    check(lambda: orig_torrent.info.force_start, not current_setting)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_set_super_seeding",
        "torrents_setSuperSeeding",
        "torrents.set_super_seeding",
        "torrents.setSuperSeeding",
    ),
)
def test_set_super_seeding(client, client_func, orig_torrent):
    # this may or may not actually be working....
    get_func(client, client_func)(enable=False, torrent_hashes=orig_torrent.hash)
    check(lambda: orig_torrent.info.force_start, False)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_add_peers",
        "torrents_addPeers",
        "torrents.add_peers",
        "torrents.addPeers",
    ),
)
@pytest.mark.parametrize(
    "peers", ("127.0.0.1:5000", ("127.0.0.1:5000", "127.0.0.2:5000"), "127.0.0.1")
)
def test_torrents_add_peers(client, api_version, orig_torrent, client_func, peers):
    if v(api_version) >= v("2.3.0"):
        if all(map(lambda p: ":" not in p, peers)):
            with pytest.raises(InvalidRequest400Error):
                get_func(client, client_func)(
                    peers=peers, torrent_hashes=orig_torrent.hash
                )
        else:
            p = get_func(client, client_func)(
                peers=peers, torrent_hashes=orig_torrent.hash
            )
            # can only actually verify the peer was added if it's a valid peer  # noqa: E800
            # check(  # noqa: E800
            #     lambda: client.sync_torrent_peers(torrent_hash=orig_torrent.hash)['peers'],  # noqa: E800
            #     peers,  # noqa: E800
            #     reverse=True  # noqa: E800
            # )  # noqa: E800
            assert isinstance(p, TorrentsAddPeersDictionary)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(peers=peers, torrent_hashes=orig_torrent.hash)


def _categories_save_path_key(api_version):
    """With qBittorrent 4.4.0 (Web API 2.8.4), the key in the category
    definition returned changed from savePath to save_path...."""
    if v(api_version) == v("2.8.4"):
        return "save_path"
    return "savePath"


def test_categories1(client, api_version):
    if v(api_version) >= v("2.1.1"):
        assert isinstance(client.torrents_categories(), TorrentCategoriesDictionary)
    else:
        with pytest.raises(NotImplementedError):
            client.torrents_categories()


def test_categories2(client, api_version):
    if v(api_version) >= v("2.1.1"):
        save_path_key = _categories_save_path_key(api_version)
        name = "new_category"
        client.torrent_categories.categories = {"name": name, save_path_key: "/tmp"}
        assert name in client.torrent_categories.categories
        client.torrent_categories.categories = {"name": name, save_path_key: "/tmp/new"}
        assert mkpath(
            client.torrent_categories.categories[name][save_path_key]
        ) == mkpath("/tmp/new")
        client.torrents_remove_categories(categories=name)
    else:
        with pytest.raises(NotImplementedError):
            client.torrent_categories.categories


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_create_category",
        "torrents_createCategory",
        "torrent_categories.create_category",
        "torrent_categories.createCategory",
    ),
)
@pytest.mark.parametrize("filepath", (None, "", "/tmp/"))
@pytest.mark.parametrize("name", ("name", "name 1"))
@pytest.mark.parametrize("enable_download_path", (None, True, False))
def test_create_categories(
    client, api_version, orig_torrent, client_func, filepath, name, enable_download_path
):
    save_path = download_path = filepath
    if filepath:
        save_path += "save"
        download_path += "download"

    try:
        get_func(client, client_func)(
            name=name,
            save_path=save_path,
            download_path=download_path,
            enable_download_path=enable_download_path,
        )
        client.torrents_set_category(torrent_hashes=orig_torrent.hash, category=name)
        check(lambda: orig_torrent.info.category.replace("+", " "), name)
        if v(api_version) >= v("2.2"):
            check(
                lambda: [n.replace("+", " ") for n in client.torrents_categories()],
                name,
                reverse=True,
            )
            save_path_key = _categories_save_path_key(api_version)
            check(
                lambda: [
                    mkpath(cat[save_path_key])
                    for cat in client.torrents_categories().values()
                ],
                mkpath(save_path) or "",
                reverse=True,
            )
        if v(api_version) >= v("2.8.4") and enable_download_path is not False:
            check(
                lambda: [
                    mkpath(cat.get("download_path", ""))
                    for cat in client.torrents_categories().values()
                ],
                mkpath(download_path) or "",
                reverse=True,
            )
    finally:
        client.torrents_remove_categories(categories=name)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_edit_category",
        "torrents_editCategory",
        "torrent_categories.edit_category",
        "torrent_categories.editCategory",
    ),
)
@pytest.mark.parametrize("filepath", ("", "/tmp/"))
@pytest.mark.parametrize("name", ("editcategory",))
@pytest.mark.parametrize("enable_download_path", (None, True, False))
def test_edit_category(
    client, api_version, client_func, filepath, name, enable_download_path
):
    if v(api_version) >= v("2.1.0") and filepath is not None:
        try:
            client.torrents_create_category(
                name=name, save_path="/tmp/savetmp", download_path="/tmp/savetmp"
            )
            save_path = mkpath(filepath + "save/")
            download_path = mkpath(filepath + "down/")
            get_func(client, client_func)(
                name=name,
                save_path=save_path,
                download_path=download_path,
                enable_download_path=enable_download_path,
            )
            check(
                lambda: [n.replace("+", " ") for n in client.torrents_categories()],
                name,
                reverse=True,
            )
            save_path_key = _categories_save_path_key(api_version)
            check(
                lambda: (
                    mkpath(cat[save_path_key])
                    for cat in client.torrents_categories().values()
                ),
                mkpath(save_path) or "",
                reverse=True,
            )
            if v(api_version) >= v("2.8.4") and enable_download_path is not False:
                check(
                    lambda: [
                        mkpath(cat.get("download_path", ""))
                        for cat in client.torrents_categories().values()
                    ],
                    mkpath(download_path) or "",
                    reverse=True,
                )
        finally:
            client.torrents_remove_categories(categories=name)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(name=name, save_path=filepath)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_remove_categories",
        "torrents_removeCategories",
        "torrent_categories.remove_categories",
        "torrent_categories.removeCategories",
    ),
)
@pytest.mark.parametrize("categories", (("category1",), ("category1", "category 2")))
def test_remove_category(client, api_version, orig_torrent, client_func, categories):
    for name in categories:
        client.torrents_create_category(name=name)
    orig_torrent.set_category(category=categories[0])
    get_func(client, client_func)(categories=categories)
    if v(api_version) >= v("2.2"):
        check(
            lambda: [n.replace("+", " ") for n in client.torrents_categories()],
            categories,
            reverse=True,
            negate=True,
        )
    check(lambda: orig_torrent.info.category, categories[0], negate=True)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_tags",
        "torrent_tags.tags",
    ),
)
def test_tags(client, api_version, client_func):
    if v(api_version) >= v("2.3.0"):
        try:
            assert isinstance(get_func(client, client_func)(), TagList)
        except TypeError:
            assert isinstance(get_func(client, client_func), TagList)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)()


def test_add_tag_though_property(client, api_version):
    name = "newtag"
    if v(api_version) >= v("2.3.0"):
        client.torrent_tags.tags = name
        assert name in client.torrent_tags.tags
        client.torrent_tags.delete_tags(name)
        assert name not in client.torrent_tags.tags
    else:
        with pytest.raises(NotImplementedError):
            client.torrent_tags.tags = name


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_add_tags",
        "torrents_addTags",
        "torrent_tags.add_tags",
        "torrent_tags.addTags",
    ),
)
@pytest.mark.parametrize("tags", (("tag1",), ("tag1", "tag 2")))
def test_add_tags(client, api_version, orig_torrent, client_func, tags):
    if v(api_version) >= v("2.3.0"):
        try:
            get_func(client, client_func)(tags=tags, torrent_hashes=orig_torrent.hash)
            check(lambda: orig_torrent.info.tags, tags, reverse=True)
        finally:
            client.torrents_delete_tags(tags=tags)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(tags=tags, torrent_hashes=orig_torrent.hash)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_remove_tags",
        "torrents_removeTags",
        "torrent_tags.remove_tags",
        "torrent_tags.removeTags",
    ),
)
@pytest.mark.parametrize("tags", (("tag1",), ("tag1", "tag 2")))
def test_remove_tags(client, api_version, orig_torrent, client_func, tags):
    if v(api_version) >= v("2.3.0"):
        try:
            orig_torrent.add_tags(tags=tags)
            get_func(client, client_func)(tags=tags, torrent_hashes=orig_torrent.hash)
            check(lambda: orig_torrent.info.tags, tags, reverse=True, negate=True)
        finally:
            client.torrents_delete_tags(tags=tags)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(tags=tags, torrent_hashes=orig_torrent.hash)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_create_tags",
        "torrents_createTags",
        "torrent_tags.create_tags",
        "torrent_tags.createTags",
    ),
)
@pytest.mark.parametrize("tags", (("tag1",), ("tag1", "tag 2")))
def test_create_tags(client, api_version, client_func, tags):
    if v(api_version) >= v("2.3.0"):
        try:
            get_func(client, client_func)(tags=tags)
            check(lambda: client.torrents_tags(), tags, reverse=True)
        finally:
            client.torrents_delete_tags(tags=tags)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(tags=tags)


@pytest.mark.parametrize(
    "client_func",
    (
        "torrents_delete_tags",
        "torrents_deleteTags",
        "torrent_tags.delete_tags",
        "torrent_tags.deleteTags",
    ),
)
@pytest.mark.parametrize("tags", (("tag1",), ("tag1", "tag 2")))
def test_delete_tags(client, api_version, client_func, tags):
    if v(api_version) >= v("2.3.0"):
        client.torrents_create_tags(tags=tags)
        get_func(client, client_func)(tags=tags)
        check(lambda: client.torrents_tags(), tags, reverse=True, negate=True)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)(tags=tags)



import platform
from time import sleep

import pytest
from pkg_resources import parse_version as v

from qbittorrentapi import Conflict409Error
from qbittorrentapi import TorrentDictionary
from qbittorrentapi import TorrentFilesList
from qbittorrentapi import TorrentPieceInfoList
from qbittorrentapi import TorrentPropertiesDictionary
from qbittorrentapi import TorrentStates
from qbittorrentapi import TrackersList
from qbittorrentapi import WebSeedsList
from tests.conftest import get_func
from tests.test_torrents import check
from tests.test_torrents import disable_queueing
from tests.test_torrents import enable_queueing
from tests.test_torrents import mkpath
from tests.test_torrents import torrent1_hash
from tests.test_torrents import torrent1_url


def test_info(orig_torrent, monkeypatch):
    assert orig_torrent.info.hash == orig_torrent.hash
    # mimic <=v2.0.1 where torrents_info() doesn't support hash arg
    orig_torrent._client._MOCK_WEB_API_VERSION = "2"
    assert orig_torrent.info.hash == orig_torrent.hash
    orig_torrent._client._MOCK_WEB_API_VERSION = None

    # ensure if things are really broken, an empty TorrentDictionary is returned...
    if platform.python_implementation() == "CPython":
        with monkeypatch.context() as m:

            def fake_info(*args, **kwargs):
                return []

            m.setattr(orig_torrent._client, "torrents_info", fake_info)
            assert isinstance(orig_torrent.info, TorrentDictionary)
            assert bool(orig_torrent.info) is False


def test_sync_local(orig_torrent):
    orig_torrent.state = "gibberish"
    assert orig_torrent.state == "gibberish"
    orig_torrent.sync_local()
    assert isinstance(orig_torrent, TorrentDictionary)
    assert orig_torrent.state != "gibberish"


def test_state_enum(orig_torrent):
    assert orig_torrent.state_enum in TorrentStates
    assert orig_torrent.state_enum is not TorrentStates.UNKNOWN
    check(lambda: orig_torrent.state_enum.is_downloading, True)
    # simulate an unknown torrent.state
    orig_torrent.state = "gibberish"
    assert orig_torrent.state_enum is TorrentStates.UNKNOWN
    # restore torrent state
    orig_torrent.sync_local()
    check(lambda: orig_torrent.state_enum.is_downloading, True)


def test_pause_resume(client, orig_torrent):
    orig_torrent.pause()
    check(
        lambda: client.torrents_info(torrents_hashes=orig_torrent.hash)[0].state,
        ("stalledDL", "pausedDL"),
        any=True,
    )

    orig_torrent.resume()
    check(
        lambda: client.torrents_info(torrents_hashes=orig_torrent.hash)[0].state,
        ("pausedDL",),
        negate=True,
        check_limit=20,
    )


def test_delete(client):
    client.torrents_add(urls=torrent1_url)
    check(lambda: [t.hash for t in client.torrents_info()], torrent1_hash, reverse=True)
    torrent = [t for t in client.torrents_info() if t.hash == torrent1_hash][0]
    sleep(1)
    torrent.delete(delete_files=True)
    check(
        lambda: [t.hash for t in client.torrents_info()],
        torrent1_hash,
        reverse=True,
        negate=True,
    )


@pytest.mark.parametrize(
    "client_func",
    (
        ("increase_priority", "decrease_priority", "top_priority", "bottom_priority"),
        ("increasePrio", "decreasePrio", "topPrio", "bottomPrio"),
    ),
)
def test_priority(client, new_torrent, client_func):
    disable_queueing(client)

    with pytest.raises(Conflict409Error):
        get_func(new_torrent, client_func[0])()
    with pytest.raises(Conflict409Error):
        get_func(new_torrent, client_func[1])()
    with pytest.raises(Conflict409Error):
        get_func(new_torrent, client_func[2])()
    with pytest.raises(Conflict409Error):
        get_func(new_torrent, client_func[3])()

    enable_queueing(client)
    sleep(1)  # putting sleeps in since these keep crashing qbittorrent

    current_priority = new_torrent.info.priority
    get_func(new_torrent, client_func[0])()
    sleep(1)
    check(lambda: new_torrent.info.priority < current_priority, True)

    current_priority = new_torrent.info.priority
    get_func(new_torrent, client_func[1])()
    sleep(1)
    check(lambda: new_torrent.info.priority > current_priority, True)

    current_priority = new_torrent.info.priority
    get_func(new_torrent, client_func[2])()
    sleep(1)
    check(lambda: new_torrent.info.priority < current_priority, True)

    current_priority = new_torrent.info.priority
    get_func(new_torrent, client_func[3])()
    sleep(1)
    check(lambda: new_torrent.info.priority > current_priority, True)


@pytest.mark.parametrize("client_func", ("set_share_limits", "setShareLimits"))
def test_set_share_limits(api_version, orig_torrent, client_func):
    if v(api_version) >= v("2.0.1"):
        get_func(orig_torrent, client_func)(ratio_limit=5, seeding_time_limit=100)
        check(lambda: orig_torrent.info.max_ratio, 5)
        check(lambda: orig_torrent.info.max_seeding_time, 100)
    else:
        with pytest.raises(NotImplementedError):
            get_func(orig_torrent, client_func)(ratio_limit=5, seeding_time_limit=100)


@pytest.mark.parametrize(
    "client_func",
    (("download_limit", "set_download_limit"), ("downloadLimit", "setDownloadLimit")),
)
def test_download_limit(orig_torrent, client_func):
    setattr(orig_torrent, client_func[0], 2048)
    check(lambda: get_func(orig_torrent, client_func[0]), 2048)
    check(lambda: orig_torrent.info.dl_limit, 2048)

    get_func(orig_torrent, client_func[1])(4096)
    check(lambda: get_func(orig_torrent, client_func[0]), 4096)
    check(lambda: orig_torrent.info.dl_limit, 4096)


@pytest.mark.parametrize(
    "client_func",
    (("upload_limit", "set_upload_limit"), ("uploadLimit", "setUploadLimit")),
)
def test_upload_limit(orig_torrent, client_func):
    setattr(orig_torrent, client_func[0], 2048)
    check(lambda: get_func(orig_torrent, client_func[0]), 2048)
    check(lambda: orig_torrent.info.up_limit, 2048)

    get_func(orig_torrent, client_func[1])(4096)
    check(lambda: get_func(orig_torrent, client_func[0]), 4096)
    check(lambda: orig_torrent.info.up_limit, 4096)


@pytest.mark.parametrize("client_func", ("set_location", "setLocation"))
def test_set_location(api_version, new_torrent, client_func):
    if v(api_version) >= v("2.0.2"):
        exp = None
        for attempt in range(2):
            try:
                loc = mkpath("/tmp", "3")
                get_func(new_torrent, client_func)(loc)
                check(
                    lambda: mkpath(new_torrent.info.save_path),
                    mkpath(loc),
                    any=True,
                )
                break
            except AssertionError as e:
                exp = e
        if exp:
            raise exp


@pytest.mark.parametrize("client_func", ("set_save_path", "setSavePath"))
def test_set_save_path(api_version, new_torrent, client_func):
    if v(api_version) >= v("2.8.4"):
        exp = None
        for attempt in range(2):
            try:
                loc = mkpath("/tmp", "savepath3")
                get_func(new_torrent, client_func)(loc)
                # qBittorrent may return trailing separators depending on version....
                check(
                    lambda: mkpath(new_torrent.info.save_path),
                    mkpath(loc),
                    any=True,
                )
                break
            except AssertionError as e:
                exp = e
        if exp:
            raise exp


@pytest.mark.parametrize("client_func", ("set_download_path", "setDownloadPath"))
def test_set_download_path(api_version, new_torrent, client_func):
    if v(api_version) >= v("2.8.4"):
        exp = None
        for attempt in range(2):
            try:
                loc = mkpath("/tmp", "downloadpath3")
                get_func(new_torrent, client_func)(loc)
                # qBittorrent may return trailing separators depending on version....
                check(
                    lambda: mkpath(new_torrent.info.download_path),
                    mkpath(loc),
                    any=True,
                )
                break
            except AssertionError as e:
                exp = e
        if exp:
            raise exp


@pytest.mark.parametrize("client_func", ("set_category", "setCategory"))
@pytest.mark.parametrize("category", ("category 1", "category_1"))
def test_set_category(client, orig_torrent, client_func, category):
    client.torrents_create_category(category=category)
    get_func(orig_torrent, client_func)(category=category)
    check(lambda: orig_torrent.info.category.replace("+", " "), category, reverse=True)
    client.torrents_remove_categories(categories=category)


@pytest.mark.parametrize("client_func", ("set_auto_management", "setAutoManagement"))
def test_set_auto_management(orig_torrent, client_func):
    current_setting = orig_torrent.auto_tmm
    get_func(orig_torrent, client_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.auto_tmm, not current_setting)
    get_func(orig_torrent, client_func)(enable=current_setting)
    check(lambda: orig_torrent.info.auto_tmm, current_setting)


@pytest.mark.parametrize(
    "client_func", ("toggle_sequential_download", "toggleSequentialDownload")
)
def test_toggle_sequential_download(orig_torrent, client_func):
    current_setting = orig_torrent.seq_dl
    get_func(orig_torrent, client_func)()
    check(lambda: orig_torrent.info.seq_dl, not current_setting)
    get_func(orig_torrent, client_func)()
    check(lambda: orig_torrent.info.seq_dl, current_setting)


@pytest.mark.parametrize(
    "client_func", ("toggle_first_last_piece_priority", "toggleFirstLastPiecePrio")
)
def test_toggle_first_last_piece_priority(api_version, orig_torrent, client_func):
    if v(api_version) >= v("2.0.2"):
        current_setting = orig_torrent.f_l_piece_prio
        get_func(orig_torrent, client_func)()
        sleep(1)
        check(lambda: orig_torrent.info.f_l_piece_prio, not current_setting)


@pytest.mark.parametrize("client_func", ("set_force_start", "setForceStart"))
def test_set_force_start(orig_torrent, client_func):
    current_setting = orig_torrent.force_start
    get_func(orig_torrent, client_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.force_start, not current_setting)
    get_func(orig_torrent, client_func)(enable=current_setting)
    check(lambda: orig_torrent.info.force_start, current_setting)


@pytest.mark.parametrize("client_func", ("set_super_seeding", "setSuperSeeding"))
def test_set_super_seeding(orig_torrent, client_func):
    current_setting = orig_torrent.super_seeding
    get_func(orig_torrent, client_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.super_seeding, not current_setting)
    get_func(orig_torrent, client_func)(enable=current_setting)
    check(lambda: orig_torrent.info.super_seeding, current_setting)


def test_properties(orig_torrent):
    assert isinstance(orig_torrent.properties, TorrentPropertiesDictionary)
    assert "save_path" in orig_torrent.properties


@pytest.mark.parametrize("trackers", ("127.0.0.2", ("127.0.0.3", "127.0.0.4")))
def test_trackers(orig_torrent, trackers):
    assert isinstance(orig_torrent.trackers, TrackersList)
    assert "num_peers" in orig_torrent.trackers[-1]

    orig_torrent.trackers = trackers
    check(lambda: (t.url for t in orig_torrent.trackers), trackers, reverse=True)


@pytest.mark.parametrize("client_func", ("add_trackers", "addTrackers"))
@pytest.mark.parametrize("trackers", ("127.0.0.2", ("127.0.0.3", "127.0.0.4")))
def test_add_tracker(new_torrent, client_func, trackers):
    get_func(new_torrent, client_func)(urls=trackers)
    sleep(0.5)  # try to stop crashing qbittorrent
    check(lambda: (t.url for t in new_torrent.trackers), trackers, reverse=True)


@pytest.mark.parametrize("client_func", ("edit_tracker", "editTracker"))
def test_edit_tracker(api_version, orig_torrent, client_func):
    if v(api_version) >= v("2.2.0"):
        orig_torrent.add_trackers(urls="127.0.1.1")
        get_func(orig_torrent, client_func)(orig_url="127.0.1.1", new_url="127.0.1.2")
        check(
            lambda: (t.url for t in orig_torrent.trackers),
            "127.0.1.1",
            reverse=True,
            negate=True,
        )
        check(lambda: (t.url for t in orig_torrent.trackers), "127.0.1.2", reverse=True)
        get_func(orig_torrent, "remove_trackers")(urls="127.0.1.2")
    else:
        with pytest.raises(NotImplementedError):
            get_func(orig_torrent, client_func)(
                orig_url="127.0.1.1", new_url="127.0.1.2"
            )


@pytest.mark.parametrize("client_func", ("remove_trackers", "removeTrackers"))
@pytest.mark.parametrize("trackers", ("127.0.2.2", ("127.0.2.3", "127.0.2.4")))
def test_remove_trackers(api_version, orig_torrent, client_func, trackers):
    if v(api_version) >= v("2.2.0"):
        check(
            lambda: (t.url for t in orig_torrent.trackers),
            trackers,
            reverse=True,
            negate=True,
        )
        orig_torrent.add_trackers(urls=trackers)
        check(lambda: (t.url for t in orig_torrent.trackers), trackers, reverse=True)
        get_func(orig_torrent, client_func)(urls=trackers)
        check(
            lambda: (t.url for t in orig_torrent.trackers),
            trackers,
            reverse=True,
            negate=True,
        )
    else:
        with pytest.raises(NotImplementedError):
            get_func(orig_torrent, client_func)(urls=trackers)


def test_webseeds(orig_torrent):
    assert isinstance(orig_torrent.webseeds, WebSeedsList)


def test_files(orig_torrent):
    assert isinstance(orig_torrent.files, TorrentFilesList)
    assert "id" in orig_torrent.files[0]


def test_recheck(orig_torrent):
    orig_torrent.recheck()


def test_reannounce(api_version, orig_torrent):
    if v(api_version) >= v("2.0.2"):
        orig_torrent.reannounce()
    else:
        with pytest.raises(NotImplementedError):
            orig_torrent.reannounce()


@pytest.mark.parametrize("client_func", ("rename_file", "renameFile"))
@pytest.mark.parametrize("name", ("new_name", "new name"))
def test_rename_file(api_version, app_version, new_torrent, client_func, name):
    sleep(1)
    if v(api_version) >= v("2.4.0"):
        get_func(new_torrent, client_func)(file_id=0, new_file_name=name)
        check(lambda: new_torrent.files[0].name, name)
    else:
        with pytest.raises(NotImplementedError):
            get_func(new_torrent, client_func)(file_id=0, new_file_name=name)

    if v(app_version) >= v("v4.3.3"):
        curr_name = new_torrent.files[0].name
        new_name = "NEW_" + name
        get_func(new_torrent, client_func)(old_path=curr_name, new_path=new_name)
        check(lambda: new_torrent.files[0].name, new_name)


@pytest.mark.parametrize("client_func", ("rename_folder", "renameFolder"))
@pytest.mark.parametrize("name", ("new_name", "new name"))
def test_rename_folder(api_version, app_version, new_torrent, client_func, name):
    # need to ensure we're at least on v4.3.3 to run test since
    # both v4.3.2 and v4.3.2 both use Web API 2.7
    if v(api_version) >= v("2.7") and v(app_version) >= v("v4.3.3"):
        # move the file in to a new folder
        orig_file_path = new_torrent.files[0].name
        new_folder = "qwer"
        new_torrent.rename_file(
            old_path=orig_file_path,
            new_path=new_folder + "/" + orig_file_path,
        )
        sleep(1)  # qBittorrent crashes if you make these calls too fast...
        # test rename that new folder
        get_func(new_torrent, client_func)(
            old_path=new_folder,
            new_path=name,
        )
        check(
            lambda: new_torrent.files[0].name.replace("+", " "),
            name + "/" + orig_file_path,
        )
    else:
        with pytest.raises(NotImplementedError):
            get_func(new_torrent, client_func)(old_path="", new_path="")


@pytest.mark.parametrize("client_func", ("piece_states", "pieceStates"))
def test_piece_states(orig_torrent, client_func):
    assert isinstance(get_func(orig_torrent, client_func), TorrentPieceInfoList)


@pytest.mark.parametrize("client_func", ("piece_hashes", "pieceHashes"))
def test_piece_hashes(orig_torrent, client_func):
    assert isinstance(get_func(orig_torrent, client_func), TorrentPieceInfoList)


@pytest.mark.parametrize("client_func", ("file_priority", "filePriority"))
def test_file_priority(orig_torrent, client_func):
    get_func(orig_torrent, client_func)(file_ids=0, priority=7)
    check(lambda: orig_torrent.files[0].priority, 7)


@pytest.mark.parametrize("name", ("new_name", "new name"))
def test_rename(orig_torrent, name):
    orig_torrent.rename(new_name=name)
    check(lambda: orig_torrent.info.name.replace("+", " "), name)


@pytest.mark.parametrize(
    "client_func", (("add_tags", "remove_tags"), ("addTags", "removeTags"))
)
@pytest.mark.parametrize("tags", ("tag 1", ("tag 2", "tag 3")))
def test_add_remove_tags(client, api_version, orig_torrent, client_func, tags):
    if v(api_version) >= v("2.3.0"):
        get_func(orig_torrent, client_func[0])(tags=tags)
        check(lambda: orig_torrent.info.tags, tags, reverse=True)

        get_func(orig_torrent, client_func[1])(tags=tags)
        check(lambda: orig_torrent.info.tags, tags, reverse=True, negate=True)

        client.torrents_delete_tags(tags=tags)
    else:
        with pytest.raises(NotImplementedError):
            get_func(orig_torrent, client_func[0])(tags=tags)
        with pytest.raises(NotImplementedError):
            get_func(orig_torrent, client_func[1])(tags=tags)



from enum import Enum

import pytest

from qbittorrentapi._attrdict import AttrDict
from qbittorrentapi.definitions import Dictionary
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry
from qbittorrentapi.definitions import TorrentStates

all_states = (
    "error",
    "missingFiles",
    "uploading",
    "pausedUP",
    "queuedUP",
    "stalledUP",
    "checkingUP",
    "forcedUP",
    "allocating",
    "downloading",
    "metaDL",
    "forcedMetaDL",
    "pausedDL",
    "queuedDL",
    "forcedDL",
    "stalledDL",
    "checkingDL",
    "checkingResumeData",
    "moving",
    "unknown",
)

downloading_states = (
    "downloading",
    "metaDL",
    "forcedMetaDL",
    "stalledDL",
    "checkingDL",
    "pausedDL",
    "queuedDL",
    "forcedDL",
)

uploading_states = ("uploading", "stalledUP", "checkingUP", "queuedUP", "forcedUP")

complete_states = (
    "uploading",
    "stalledUP",
    "checkingUP",
    "pausedUP",
    "queuedUP",
    "forcedUP",
)

checking_states = ("checkingUP", "checkingDL", "checkingResumeData")

errored_states = ("missingFiles", "error")

paused_states = ("pausedUP", "pausedDL")


def test_torrent_states_exists():
    assert isinstance(TorrentStates("unknown"), Enum)


@pytest.mark.parametrize("state", all_states)
def test_all_states(state):
    assert TorrentStates(state) in TorrentStates


@pytest.mark.parametrize("state", downloading_states)
def test_downloading_states(state):
    assert TorrentStates(state).is_downloading


@pytest.mark.parametrize("state", uploading_states)
def test_uploading_states(state):
    assert TorrentStates(state).is_uploading


@pytest.mark.parametrize("state", complete_states)
def test_completing_states(state):
    assert TorrentStates(state).is_complete


@pytest.mark.parametrize("state", checking_states)
def test_checking_states(state):
    assert TorrentStates(state).is_checking


@pytest.mark.parametrize("state", errored_states)
def test_errored_states(state):
    assert TorrentStates(state).is_errored


@pytest.mark.parametrize("state", paused_states)
def test_paused_states(state):
    assert TorrentStates(state).is_paused


def test_testing_groups():
    assert set(all_states) == {s.value for s in iter(TorrentStates)}
    assert set(downloading_states) == {
        s.value for s in TorrentStates if s.is_downloading
    }
    assert set(uploading_states) == {s.value for s in TorrentStates if s.is_uploading}
    assert set(complete_states) == {s.value for s in TorrentStates if s.is_complete}
    assert set(checking_states) == {s.value for s in TorrentStates if s.is_checking}
    assert set(errored_states) == {s.value for s in TorrentStates if s.is_errored}


def test_dictionary():
    assert len(Dictionary()) == 0
    assert len(Dictionary({"one": {"two": 2}})) == 1
    assert isinstance(Dictionary({"one": {"two": 2}}).one, AttrDict)
    assert isinstance(Dictionary({"one": {"two": 2}})["one"], AttrDict)
    assert Dictionary({"one": {"two": 2}}).one.two == 2
    assert Dictionary({"one": {"two": 2}})["one"]["two"] == 2
    assert Dictionary({"three": 3}).three == 3
    assert Dictionary({"three": 3})["three"] == 3


def test_list():
    assert len(List()) == 0
    list_entries = ({"one": "1"}, {"two": "2"}, {"three": "3"})
    test_list = List(list_entries=list_entries, entry_class=ListEntry)
    assert len(test_list) == 3
    assert issubclass(type(test_list[0]), ListEntry)
    assert test_list[0].one == "1"



import pytest
from pkg_resources import parse_version as v

from qbittorrentapi._attrdict import AttrDict
from tests.conftest import IS_QBT_DEV


def test_version(client, app_version):
    assert client.app_version() == app_version
    assert client.app.version == app_version
    assert client.application.version == app_version


def test_web_api_version(client, api_version):
    if IS_QBT_DEV:
        return

    assert client.app_web_api_version() == api_version
    assert client.app.web_api_version == api_version
    assert client.application.web_api_version == api_version


def test_build_info(client, api_version):
    if v(api_version) >= v("2.3"):
        assert "libtorrent" in client.app_build_info()
        assert "libtorrent" in client.app.build_info
    else:
        with pytest.raises(NotImplementedError):
            assert "libtorrent" in client.app_build_info()
        with pytest.raises(NotImplementedError):
            assert "libtorrent" in client.app.build_info


def test_preferences(client):
    prefs = client.app_preferences()
    assert "dht" in prefs
    assert "dht" in client.app.preferences
    dht = prefs["dht"]
    client.app.preferences = AttrDict(dht=(not dht))
    assert dht is not client.app.preferences.dht
    client.app_set_preferences(prefs=dict(dht=dht))
    assert dht is client.app.preferences.dht


def test_default_save_path(client):
    assert "download" in client.app_default_save_path().lower()
    assert "download" in client.app.default_save_path.lower()



from functools import wraps
from json import loads
from logging import getLogger

from pkg_resources import parse_version as v

from qbittorrentapi.exceptions import APIError
from qbittorrentapi.exceptions import HTTP403Error

logger = getLogger(__name__)


class alias(object):
    """
    Alias class that can be used as a decorator for making methods callable
    through other names (or "aliases").
    Note: This decorator must be used inside an @aliased -decorated class.
    For example, if you want to make the method shout() be also callable as
    yell() and scream(), you can use alias like this:

        @alias('yell', 'scream')
        def shout(message):
            # ....
    """

    def __init__(self, *aliases):
        self.aliases = set(aliases)

    def __call__(self, func):
        """
        Method call wrapper.

        As this decorator has arguments, this method will only be called
        once as a part of the decoration process, receiving only one
        argument: the decorated function ('f'). As a result of this kind
        of decorator, this method must return the callable that will
        wrap the decorated function.
        """
        func._aliases = self.aliases
        return func


def aliased(aliased_class):
    """
    Decorator function that *must* be used in combination with @alias
    decorator. This class will make the magic happen!

    @aliased classes will have their aliased method (via @alias) actually aliased.
    This method simply iterates over the member attributes of 'aliased_class'
    seeking for those which have an '_aliases' attribute and then defines new
    members in the class using those aliases as mere pointer functions to the
    original ones.

    Usage:
        @aliased
        class MyClass(object):
            @alias('coolMethod', 'myKinkyMethod')
            def boring_method():
                # ...

        i = MyClass()
        i.coolMethod() # equivalent to i.myKinkyMethod() and i.boring_method()
    """
    original_methods = aliased_class.__dict__.copy()
    for method in original_methods.values():
        if hasattr(method, "_aliases"):
            # Add the aliases for 'method', but don't override any
            # previously-defined attribute of 'aliased_class'
            # noinspection PyProtectedMember
            for method_alias in method._aliases - set(original_methods):
                setattr(aliased_class, method_alias, method)
    return aliased_class


def login_required(func):
    """Ensure client is logged in before calling API methods."""

    def get_requests_kwargs(**kwargs):
        """Extract kwargs for performing transparent qBittorrent login."""
        return dict(
            requests_args=kwargs.get("requests_args"),
            requests_params=kwargs.get("requests_params"),
            headers=kwargs.get("headers"),
        )

    @wraps(func)
    def wrapper(client, *args, **kwargs):
        if not client.is_logged_in:
            logger.debug("Not logged in...attempting login")
            client.auth_log_in(**get_requests_kwargs(**kwargs))
        try:
            return func(client, *args, **kwargs)
        except HTTP403Error:
            logger.debug("Login may have expired...attempting new login")
            client.auth_log_in(**get_requests_kwargs(**kwargs))
            return func(client, *args, **kwargs)

    return wrapper


def handle_hashes(func):
    """
    Normalize torrent hash arguments.

    Initial implementations of this client used 'hash' and 'hashes' as
    function arguments for torrent hashes. Since 'hash' collides with an
    internal python name, all arguments were updated to 'torrent_hash'
    or 'torrent_hashes'. Since both versions of argument names remain
    respected, this decorator normalizes torrent hash arguments into
    either 'torrent_hash' or 'torrent_hashes'.
    """

    @wraps(func)
    def wrapper(client, *args, **kwargs):
        if "torrent_hash" not in kwargs and "hash" in kwargs:
            kwargs["torrent_hash"] = kwargs.pop("hash")
        elif "torrent_hashes" not in kwargs and "hashes" in kwargs:
            kwargs["torrent_hashes"] = kwargs.pop("hashes")
        return func(client, *args, **kwargs)

    return wrapper


def response_text(response_class):
    """
    Return the UTF-8 encoding of the API response.

    :param response_class: class to cast the response to
    :return: Text of the response casted to the specified class
    """

    def _inner(func):
        @wraps(func)
        def wrapper(client, *args, **kwargs):
            result = func(client, *args, **kwargs)
            if isinstance(result, response_class):
                return result
            try:
                return response_class(result.text)
            except Exception:
                logger.debug("Exception during response parsing.", exc_info=True)
                raise APIError("Exception during response parsing")

        return wrapper

    return _inner


def response_json(response_class):
    """
    Return the JSON in the API response. JSON is parsed as instance of
    response_class.

    :param response_class: class to parse the JSON in to
    :return: JSON as the response class
    """

    def _inner(func):
        @wraps(func)
        def wrapper(client, *args, **kwargs):
            simple_response = client._SIMPLE_RESPONSES or kwargs.pop(
                "SIMPLE_RESPONSES", kwargs.pop("SIMPLE_RESPONSE", False)
            )
            response = func(client, *args, **kwargs)
            try:
                if isinstance(response, response_class):
                    return response
                try:
                    result = response.json()
                except AttributeError:
                    # just in case the requests package is old and doesn't contain json()
                    result = loads(response.text)
                if simple_response:
                    return result
                return response_class(result, client)
            except Exception as exc:
                logger.debug("Exception during response parsing.", exc_info=True)
                raise APIError("Exception during response parsing. Error: %r" % exc)

        return wrapper

    return _inner


def check_for_raise(client, error_message):
    """For any nonexistent endpoint, log the error and conditionally raise an
    exception."""
    logger.debug(error_message)
    if client._RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS:
        raise NotImplementedError(error_message)


def endpoint_introduced(version_introduced, endpoint):
    """
    Prevent hitting an endpoint if the connected qBittorrent version doesn't
    support it.

    :param version_introduced: version endpoint was made available
    :param endpoint: API endpoint (e.g. /torrents/categories)
    """

    def _inner(func):
        @wraps(func)
        def wrapper(client, *args, **kwargs):

            # if the endpoint doesn't exist, return None or log an error / raise an Exception
            if v(client.app_web_api_version()) < v(version_introduced):
                error_message = (
                    "ERROR: Endpoint '%s' is Not Implemented in this version of qBittorrent. "
                    "This endpoint is available starting in Web API v%s."
                    % (endpoint, version_introduced)
                )
                check_for_raise(client=client, error_message=error_message)
                return None

            # send request to endpoint
            return func(client, *args, **kwargs)

        return wrapper

    return _inner


def version_removed(version_obsoleted, endpoint):
    """
    Prevent hitting an endpoint that was removed in a version older than the
    connected qBittorrent.

    :param version_obsoleted: the Web API version the endpoint was removed
    :param endpoint: name of the removed endpoint
    """

    def _inner(func):
        @wraps(func)
        def wrapper(client, *args, **kwargs):

            # if the endpoint doesn't exist, return None or log an error / raise an Exception
            if v(client.app_web_api_version()) >= v(version_obsoleted):
                error_message = (
                    "ERROR: Endpoint '%s' is Not Implemented. "
                    "This endpoint was removed in Web API v%s."
                    % (endpoint, version_obsoleted)
                )
                check_for_raise(client=client, error_message=error_message)
                return None

            # send request to endpoint
            return func(client, *args, **kwargs)

        return wrapper

    return _inner



from time import sleep

import pytest
from pkg_resources import parse_version as v

from qbittorrentapi import NotFound404Error
from qbittorrentapi.search import SearchJobDictionary
from qbittorrentapi.search import SearchResultsDictionary
from qbittorrentapi.search import SearchStatusesList
from tests.conftest import check
from tests.conftest import get_func

plugin_name = "legittorrents"
legit_torrents_url = "https://raw.githubusercontent.com/qbittorrent/search-plugins/master/nova3/engines/legittorrents.py"


@pytest.mark.parametrize(
    "client_func", ("search_update_plugins", "search.update_plugins")
)
def test_update_plugins(client, api_version, client_func):
    if v(api_version) >= v("2.1.1"):
        get_func(client, client_func)()
        check(
            lambda: any(
                entry["message"].startswith("Updating plugin ")
                or entry["message"] == "All plugins are already up to date."
                for entry in reversed(client.log.main())
            ),
            True,
        )
    else:
        with pytest.raises(NotImplementedError):
            client.search_update_plugins()


@pytest.mark.parametrize(
    "client_func",
    (
        ("search_plugins", "search_enable_plugin"),
        ("search.plugins", "search.enable_plugin"),
    ),
)
def test_enable_plugin(client, api_version, client_func):
    if v(api_version) >= v("2.1.1"):
        loop_limit = 3
        for loop_count in range(loop_limit):
            try:
                try:
                    plugins = get_func(client, client_func[0])()
                except TypeError:
                    plugins = get_func(client, client_func[0])
                get_func(client, client_func[1])(
                    plugins=(p["name"] for p in plugins), enable=False
                )
                check(
                    lambda: (p["enabled"] for p in client.search_plugins()),
                    True,
                    reverse=True,
                    negate=True,
                )
                get_func(client, client_func[1])(
                    plugins=(p["name"] for p in plugins), enable=True
                )
                check(
                    lambda: (p["enabled"] for p in client.search_plugins()),
                    False,
                    reverse=True,
                    negate=True,
                )
                break
            except Exception as e:
                if loop_count >= (loop_limit - 1):
                    raise e
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func[1])()


@pytest.mark.parametrize(
    "client_func",
    (
        ("search_install_plugin", "search_uninstall_plugin"),
        ("search.install_plugin", "search.uninstall_plugin"),
    ),
)
def test_install_uninstall_plugin(client, api_version, client_func):
    if v(api_version) >= v("2.1.1"):
        for _ in range(3):
            try:
                get_func(client, client_func[0])(sources=legit_torrents_url)
                check(
                    lambda: (p.name for p in client.search.plugins),
                    plugin_name,
                    reverse=True,
                )
                break
            except AssertionError:
                pass
        for _ in range(3):
            try:
                get_func(client, client_func[1])(names=plugin_name)
                check(
                    lambda: (p.name for p in client.search.plugins),
                    plugin_name,
                    reverse=True,
                    negate=True,
                )
            except AssertionError:
                pass
    else:
        with pytest.raises(NotImplementedError):
            client.search_install_plugin()
        with pytest.raises(NotImplementedError):
            client.search_uninstall_plugin()


@pytest.mark.parametrize("client_func", ("search_categories", "search.categories"))
def test_categories(client, api_version, client_func):
    if v("2.6") > v(api_version) >= v("2.1.1"):
        check(lambda: get_func(client, client_func)(), "All categories", reverse=True)
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func)()


@pytest.mark.parametrize(
    "client_func",
    (
        (
            "search_start",
            "search_status",
            "search_results",
            "search_stop",
            "search_delete",
        ),
        (
            "search.start",
            "search.status",
            "search.results",
            "search.stop",
            "search.delete",
        ),
    ),
)
def test_search(client, api_version, client_func):
    if v(api_version) >= v("2.1.1"):
        job = get_func(client, client_func[0])(
            pattern="Ubuntu", plugins="enabled", category="all"
        )
        assert isinstance(job, SearchJobDictionary)
        statuses = get_func(client, client_func[1])(search_id=job["id"])
        assert isinstance(statuses, SearchStatusesList)
        assert statuses[0]["status"] == "Running"
        results = get_func(client, client_func[2])(search_id=job["id"], limit=1)
        assert isinstance(results, SearchResultsDictionary)
        results = job.results()
        assert isinstance(results, SearchResultsDictionary)
        get_func(client, client_func[3])(search_id=job["id"])
        check(
            lambda: get_func(client, client_func[1])(search_id=job["id"])[0]["status"],
            "Stopped",
        )
        get_func(client, client_func[4])(search_id=job["id"])
        statuses = get_func(client, client_func[1])()
        assert not statuses
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func[0])()


@pytest.mark.parametrize(
    "client_func", (("search_stop", "search_start"), ("search.stop", "search.start"))
)
def test_stop(client, api_version, client_func):
    if v(api_version) >= v("2.1.1"):
        job = get_func(client, client_func[1])(
            pattern="Ubuntu", plugins="enabled", category="all"
        )
        sleep(1)
        get_func(client, client_func[0])(search_id=job.id)
        check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Stopped")

        job = get_func(client, client_func[1])(
            pattern="Ubuntu", plugins="enabled", category="all"
        )
        sleep(1)
        job.stop()
        check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Stopped")
    else:
        with pytest.raises(NotImplementedError):
            get_func(client, client_func[0])(search_id=100)


def test_delete(client, api_version):
    if v(api_version) >= v("2.1.1"):
        job = client.search_start(pattern="Ubuntu", plugins="enabled", category="all")
        job.delete()
        with pytest.raises(NotFound404Error):
            job.status()
    else:
        with pytest.raises(NotImplementedError):
            client.search_stop(search_id=100)



from sys import version_info

from qbittorrentapi import Version
from tests.conftest import IS_QBT_DEV
from tests.conftest import api_version_map


def test_supported_versions(app_version, api_version):
    if IS_QBT_DEV:
        return

    assert isinstance(Version.supported_api_versions(), set)
    assert api_version in Version.supported_api_versions()
    assert isinstance(Version.supported_app_versions(), set)
    assert app_version in Version.supported_app_versions()


def test_is_supported(app_version, api_version):
    if IS_QBT_DEV:
        return

    assert Version.is_app_version_supported(app_version) is True
    assert Version.is_app_version_supported(app_version[1:]) is True
    assert Version.is_app_version_supported("0.0.0") is False

    assert Version.is_api_version_supported(api_version) is True
    assert Version.is_api_version_supported("v" + api_version) is True
    assert Version.is_api_version_supported("0.0.0") is False


def test_latest_version():
    if IS_QBT_DEV:
        return

    # order of dictionary keys was guaranteed starting in python 3.7
    if version_info.major >= 3 and version_info.minor >= 7:
        expected_latest_api_version = list(api_version_map.values())[-1]
        expected_latest_app_version = list(api_version_map.keys())[-1]
        assert Version.latest_supported_api_version() == expected_latest_api_version
        assert Version.latest_supported_app_version() == expected_latest_app_version



from time import sleep

import pytest
from pkg_resources import parse_version as v

from qbittorrentapi.rss import RSSitemsDictionary
from tests.conftest import check
from tests.conftest import get_func

folder_one = "testFolderOne"
folder_two = "testFolderTwo"

item_one = "DistroWatch"
item_two = "DistroWatchNew"
url = "https://distrowatch.com/news/torrents.xml"


def test_refresh_item(client, api_version, rss_feed):
    if v(api_version) >= v("2.2"):
        client.rss_refresh_item(item_path=rss_feed)
        check(
            lambda: client.rss_items(include_feed_data=True)[rss_feed]["lastBuildDate"],
            "",
            negate=True,
            check_limit=20,
        )
        last_refresh = client.rss_items(include_feed_data=True)[rss_feed][
            "lastBuildDate"
        ]
        sleep(1)
        client.rss_refresh_item(item_path=rss_feed)
        check(
            lambda: client.rss_items(include_feed_data=True)[rss_feed]["lastBuildDate"],
            last_refresh,
            negate=True,
            check_limit=20,
        )
    else:
        with pytest.raises(NotImplementedError):
            client.rss_refresh_item(item_path=rss_feed)

    if v(api_version) >= v("2.2"):
        client.rss.refresh_item(item_path=rss_feed)
        check(
            lambda: client.rss_items(include_feed_data=True)[rss_feed]["lastBuildDate"],
            "",
            negate=True,
            check_limit=20,
        )
        last_refresh = client.rss_items(include_feed_data=True)[rss_feed][
            "lastBuildDate"
        ]
        sleep(1)
        client.rss.refresh_item(item_path=rss_feed)
        check(
            lambda: client.rss_items(include_feed_data=True)[rss_feed]["lastBuildDate"],
            last_refresh,
            negate=True,
            check_limit=20,
        )
    else:
        with pytest.raises(NotImplementedError):
            client.rss.refresh_item(item_path=rss_feed)


def test_items(client, api_version, rss_feed):
    if v(api_version) >= v("2.2"):
        check(lambda: client.rss_items(), rss_feed, reverse=True)
        check(lambda: client.rss_items(include_feed_data=True), rss_feed, reverse=True)
        check(
            lambda: client.rss_items(include_feed_data=True)[rss_feed],
            "articles",
            reverse=True,
        )

        check(lambda: client.rss.items(), rss_feed, reverse=True)
        check(lambda: client.rss.items.without_data, rss_feed, reverse=True)
        check(lambda: client.rss.items.with_data[rss_feed], "articles", reverse=True)


def test_add_feed(client, api_version, rss_feed):
    if v(api_version) >= v("2.2"):
        if rss_feed not in client.rss_items():
            raise Exception("rss feed not found", client.rss_items())


def test_remove_feed1(client, api_version, rss_feed):
    if v(api_version) >= v("2.2"):
        client.rss_remove_item(item_path=rss_feed)
        check(lambda: client.rss_items(), rss_feed, reverse=True, negate=True)


def test_remove_feed2(client, api_version, rss_feed):
    if v(api_version) >= v("2.2"):
        client.rss.remove_item(item_path=rss_feed)
        check(lambda: client.rss_items(), rss_feed, reverse=True, negate=True)


def test_add_remove_folder(client, api_version):
    if v(api_version) >= v("2.2"):
        name = "test_isos"

        client.rss_add_folder(folder_path=name)
        check(lambda: client.rss_items(), name, reverse=True)
        client.rss_remove_item(item_path=name)
        check(lambda: client.rss_items(), name, reverse=True, negate=True)

        client.rss.add_folder(folder_path=name)
        check(lambda: client.rss.items(), name, reverse=True)
        client.rss.remove_item(item_path=name)
        check(lambda: client.rss.items(), name, reverse=True, negate=True)


def test_move(client, api_version, rss_feed):
    if v(api_version) >= v("2.2"):
        new_name = "new_loc"

        client.rss_move_item(orig_item_path=rss_feed, new_item_path=new_name)
        check(lambda: client.rss_items(), new_name, reverse=True)

        client.rss.move_item(orig_item_path=new_name, new_item_path=rss_feed)
        check(lambda: client.rss.items(), rss_feed, reverse=True)


def test_mark_as_read(client, api_version, rss_feed):
    if v(api_version) >= v("2.5.1"):
        item_id = client.rss.items.with_data[rss_feed]["articles"][0]["id"]
        client.rss_mark_as_read(item_path=rss_feed, article_id=item_id)
        check(
            lambda: client.rss.items.with_data[rss_feed]["articles"][0],
            "isRead",
            reverse=True,
        )
    else:
        with pytest.raises(NotImplementedError):
            client.rss_mark_as_read(item_path=rss_feed, article_id=1)

    if v(api_version) >= v("2.5.1"):
        item_id = client.rss.items.with_data[rss_feed]["articles"][1]["id"]
        client.rss.mark_as_read(item_path=rss_feed, article_id=item_id)
        check(
            lambda: client.rss.items.with_data[rss_feed]["articles"][1],
            "isRead",
            reverse=True,
        )
    else:
        with pytest.raises(NotImplementedError):
            client.rss.mark_as_read(item_path=rss_feed, article_id=1)


@pytest.mark.parametrize(
    "client_func",
    (
        (
            "rss_add_feed",
            "rss_set_rule",
            "rss_rules",
            "rss_rename_rule",
            "rss_matching_articles",
            "rss_remove_rule",
            "rss_remove_item",
        ),
        (
            "rss.add_feed",
            "rss.set_rule",
            "rss.rules",
            "rss.rename_rule",
            "rss.matching_articles",
            "rss.remove_rule",
            "rss.remove_item",
        ),
    ),
)
def test_rules(client, api_version, client_func, rss_feed):
    def check_for_rule(name):
        try:
            get_func(client, client_func[2])()  # rss_rules
            check(
                lambda: get_func(client, client_func[2])(), name, reverse=True
            )  # rss_rules
        except TypeError:
            check(
                lambda: get_func(client, client_func[2]), name, reverse=True
            )  # rss_rules

    if v(api_version) >= v("2.2"):
        _ = rss_feed  # reference to avoid errors; needed to load RSS feed in to qbt
        rule_name = item_one + "Rule"
        rule_name_new = rule_name + "New"
        rule_def = {"enabled": True, "affectedFeeds": url, "addPaused": True}
        try:
            # rss_set_rule
            get_func(client, client_func[1])(rule_name=rule_name, rule_def=rule_def)
            check_for_rule(rule_name)

            if v(api_version) >= v("2.6"):  # rename was broken for a bit
                get_func(client, client_func[3])(
                    orig_rule_name=rule_name, new_rule_name=rule_name_new
                )  # rss_rename_rule
                check_for_rule(rule_name_new)
            if v(api_version) >= v("2.5.1"):
                assert isinstance(
                    get_func(client, client_func[4])(rule_name=rule_name),
                    RSSitemsDictionary,
                )  # rss_matching_articles
            else:
                with pytest.raises(NotImplementedError):
                    get_func(client, client_func[4])(
                        rule_name=rule_name
                    )  # rss_matching_articles
        finally:
            get_func(client, client_func[5])(rule_name=rule_name)  # rss_remove_rule
            get_func(client, client_func[5])(rule_name=rule_name_new)  # rss_remove_rule
            check(lambda: client.rss_rules(), rule_name, reverse=True, negate=True)
            get_func(client, client_func[6])(item_path=item_one)  # rss_remove_item
            assert item_two not in client.rss_items()
            check(lambda: client.rss_items(), item_two, reverse=True, negate=True)
